"""Set of tests for auto_carver module."""

from AutoCarver.auto_carver import *
from pytest import fixture

def test_auto_carver(x_train: DataFrame, x_test_1: DataFrame) -> None:
    """Tests AutoCarver

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    x_test_1 : DataFrame
        Simulated Test DataFrame
    """

    quantitative_features = ['Quantitative', 'Discrete_Quantitative_highnan', 'Discrete_Quantitative_lownan', 'Discrete_Quantitative', 'Discrete_Quantitative_rarevalue']
    qualitative_features = ["Qualitative", "Qualitative_grouped", "Qualitative_lownan", "Qualitative_highnan", "Discrete_Qualitative_noorder", "Discrete_Qualitative_lownan_noorder", "Discrete_Qualitative_rarevalue_noorder"]
    ordinal_features = ["Qualitative_Ordinal", "Qualitative_Ordinal_lownan", "Discrete_Qualitative_highnan"]
    values_orders = {
        "Qualitative_Ordinal": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Qualitative_Ordinal_lownan": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Discrete_Qualitative_highnan" : ["1", "2", "3", "4", "5", "6", "7"],
    }



    # minimum frequency per modality
    min_freq = 0.02
    max_n_mod = 4

    # tests with 'tschuprowt' measure
    auto_carver = AutoCarver(
        quantitative_features=quantitative_features,
        qualitative_features=qualitative_features,
        ordinal_features=ordinal_features,
        values_orders=values_orders,
        min_freq=min_freq,
        max_n_mod=max_n_mod,
        sort_by='tschuprowt',
        output_dtype='str',
        copy=True,
        verbose=False,
    )
    x_discretized = auto_carver.fit_transform(x_train, x_train["quali_ordinal_target"], X_test=x_test_1, y_test=x_test_1["quali_ordinal_target"])
    x_test_discretized = auto_carver.transform(x_test_1)

    assert all(x_discretized[auto_carver.features].nunique() <= max_n_mod), "Too many values after carving of train sample"
    assert all(x_test_discretized[auto_carver.features].nunique() <= max_n_mod), "Too many values after carving of test sample"
    assert all(x_discretized[auto_carver.features].nunique() == x_test_discretized[auto_carver.features].nunique()), "More values in train or test samples"

    
    # tests with 'cramerv' measure
    auto_carver = AutoCarver(
        quantitative_features=quantitative_features,
        qualitative_features=qualitative_features,
        ordinal_features=ordinal_features,
        values_orders=values_orders,
        min_freq=min_freq,
        max_n_mod=max_n_mod,
        sort_by='cramerv',
        copy=True,
        verbose=False,
    )
    x_discretized = auto_carver.fit_transform(x_train, x_train["quali_ordinal_target"], X_test=x_test_1, y_test=x_test_1["quali_ordinal_target"])
    x_test_discretized = auto_carver.transform(x_test_1)

    assert all(x_discretized[auto_carver.features].nunique() <= max_n_mod), "Too many values after carving of train sample"
    assert all(x_test_discretized[auto_carver.features].nunique() <= max_n_mod), "Too many values after carving of test sample"
    assert all(x_discretized[auto_carver.features].nunique() == x_test_discretized[auto_carver.features].nunique()), "More values in train or test samples"

    # test that all values still are in the values_orders
    for feature in auto_carver.qualitative_features:
        fitted_values = auto_carver.values_orders[feature].values()
        init_values = x_train[feature].fillna('__NAN__').unique()
        assert all(value in fitted_values for value in init_values), feature

    # TODO: test output 'float'
    # TODO test missing values in test



"""Tools to build simple buckets out of Quantitative features
for a binary classification model.
"""

from typing import Any, Dict, List

from numpy import array, inf, linspace, nan, quantile
from pandas import DataFrame, Series, isna, notna

from .base_discretizers import GroupedList, GroupedListDiscretizer, applied_to_dict_list


class QuantileDiscretizer(GroupedListDiscretizer):
    """Builds per-feature buckets of quantiles"""

    def __init__(
        self,
        features: list[str],
        min_freq: float,
        *,
        values_orders: dict[str, Any] = None,
        str_nan: str = "__NAN__",
        copy: bool = False,
        verbose: bool = False,
    ) -> None:
        """Discretizes quantitative features into groups of q quantiles

        Parameters
        ----------
        features : List[str]
            _description_
        min_freq : float
            _description_
        values_orders : Dict[str, Any], optional
            _description_, by default None
        str_nan : str, optional
            _description_, by default '__NAN__'
        copy : bool, optional
            _description_, by default False
        verbose : bool, optional
            _description_, by default False
        """
        # Initiating GroupedListDiscretizer
        super().__init__(
            features=features,
            values_orders=values_orders,
            input_dtypes='float',
            output_dtype='str',
            str_nan=str_nan,
            copy=copy,
        )

        self.min_freq = min_freq
        self.q = round(1 / min_freq)  # number of quantiles
        self.verbose = verbose

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None
        """
        if self.verbose:  # verbose if requested
            print(f" - [QuantileDiscretizer] Fit {', '.join(self.features)}")

        # computing quantiles for the feature
        quantiles = applied_to_dict_list(X[self.features].apply(find_quantiles, q=self.q, axis=0))

        # storing ordering
        for feature in self.features:
            # Converting to a groupedlist
            order = GroupedList(quantiles[feature] + [inf])

            # adding NANs if ther are any
            if any(isna(X[feature])):
                order.append(self.str_nan)

            # storing into the values_orders
            self.values_orders.update({feature: order})

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self


def find_quantiles(
    df_feature: Series,
    q: int,
    len_df: int = None,
    quantiles: List[float] = None,
) -> List[float]:
    """[Quantitative] Découpage en quantile de la feature.

    Fonction récursive : on prend l'échantillon et on cherche des valeurs sur-représentées
    Si la valeur existe on la met dans une classe et on cherche à gauche et à droite de celle-ci, d'autres variables sur-représentées
    Une fois il n'y a plus de valeurs qui soient sur-représentées,
    on fait un découpage classique avec qcut en multipliant le nombre de classes souhaité par le pourcentage de valeurs restantes sur le total

    """

    # initialisation de la taille total du dataframe
    if len_df is None:
        len_df = len(df_feature)

    # initialisation de la liste des quantiles
    if quantiles is None:
        quantiles = []

    # calcul du nombre d'occurences de chaque valeur
    frequencies = (
        df_feature.value_counts(dropna=False, normalize=False).drop(nan, errors="ignore") / len_df
    )  # dropping nans to keep them anyways
    values, frequencies = array(frequencies.index), array(frequencies.values)

    # case 1: no observation, all values have been attributed there corresponding modality
    if len(df_feature) == 0:
        return quantiles

    # case 2: there are missing values
    if any(isna(df_feature)):
        return find_quantiles(
            df_feature[notna(df_feature)],
            q,
            len_df=len_df,
            quantiles=quantiles,
        )

    # case 3 : there are no missing values
    # case 3.1 : there is an over-populated value
    if any(frequencies > 1 / q):
        # identification de la valeur sur-représentée
        frequent_value = values[frequencies.argmax()]

        # ajout de la valeur fréquente à la liste des quantiles
        quantiles += [frequent_value]

        # calcul des quantiles pour les parties inférieures et supérieures
        quantiles_inf = find_quantiles(df_feature[df_feature < frequent_value], q, len_df=len_df)
        quantiles_sup = find_quantiles(df_feature[df_feature > frequent_value], q, len_df=len_df)

        return quantiles_inf + quantiles + quantiles_sup

    # case 3.2 : there is no over-populated value
    # nouveau nombre de quantile en prenant en compte les classes déjà constituées
    new_q = max(round(len(df_feature) / len_df * q), 1)

    # calcul des quantiles sur le dataframe
    if new_q > 1:
        quantiles += list(
            quantile(
                df_feature.values,
                linspace(0, 1, new_q + 1)[1:-1],
                method="lower",
            )
        )

    # case when there are no enough observations to compute quantiles
    else:
        quantiles += [max(df_feature.values)]

    return quantiles



"""Tools to build simple buckets out of Qualitative features
for a binary classification model.
"""

from typing import Any, Dict, List, Union
from AutoCarver.discretizers.utils.base_discretizers import GroupedList

from numpy import argmin, nan, select
from pandas import DataFrame, Series, isna, notna, unique

from .base_discretizers import (
    GroupedList,
    GroupedListDiscretizer,
    check_missing_values,
    check_new_values,
    convert_to_labels,
    convert_to_values,
    nan_unique,
    target_rate,
    value_counts,
)
from .type_discretizers import StringDiscretizer

class DefaultDiscretizer(GroupedListDiscretizer):
    """Groups a qualitative features' values less frequent than min_freq into a str_default string

    Only use for qualitative non-ordinal features
    """

    def __init__(
        self,
        features: List[str],
        min_freq: float,
        *,
        values_orders: Dict[str, GroupedList] = None,
        str_default: str = "__OTHER__",
        str_nan: str = "__NAN__",
        copy: bool = False,
        verbose: bool = False,
    ) -> None:
        """_summary_

        Parameters
        ----------
        features : List[str]
            List of column names to be discretized
        min_freq : float
            Minimum frequency per modality. Less frequent modalities are grouped in the closest value of the order.
        values_orders : Dict[str, GroupedList], optional
            Dict of column names (keys) and modalities' associated order (values), by default None
        str_default : str, optional
            _description_, by default "__OTHER__"
        str_nan : str, optional
            _description_, by default "__NAN__"
        copy : bool, optional
            _description_, by default False
        verbose : bool, optional
            _description_, by default False
        """
        # Initiating GroupedListDiscretizer
        super().__init__(
            features=features,
            values_orders=values_orders,
            input_dtypes='str',
            output_dtype='str',
            str_nan=str_nan,
            copy=copy,
        )
        
        self.min_freq = min_freq
        self.str_default = str_default
        self.verbose = verbose

    def prepare_data(self, X: DataFrame, y: Series) -> DataFrame:
        """Called during fit step

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_

        Returns
        -------
        DataFrame
            _description_
        """
        # checking for binary target
        x_copy = super().prepare_data(X, y)

        # checks and initilizes values_orders
        for feature in self.features:
            # initiating features missing from values_orders
            if feature not in self.values_orders:
                self.values_orders.update({feature: GroupedList(nan_unique(x_copy[feature]))})

        # checking that all unique values in X are in values_orders
        check_new_values(x_copy, self.features, self.values_orders)
        # checking that all unique values in values_orders are in X
        check_missing_values(x_copy, self.features, self.values_orders)

        # adding NANS
        for feature in self.features:
            if any(x_copy[feature].isna()):
                self.values_orders[feature].append(self.str_nan)

        # filling up NaNs
        x_copy = x_copy[self.features].fillna(self.str_nan)

        return x_copy

    def fit(self, X: DataFrame, y: Series) -> None:
        """Learns modalities that are common enough (greater than min_freq)

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_
        """
        # copying dataframe and checking data before bucketization
        x_copy = self.prepare_data(X, y)
        
        if self.verbose:  # verbose if requested
            print(f" - [DefaultDiscretizer] Fit {', '.join(self.features)}")

        # computing frequencies of each modality
        frequencies = x_copy.apply(value_counts, normalize=True, axis=0)

        for feature in self.features:
            # sorting orders based on target rates
            order = self.values_orders[feature]

            # checking for rare values
            values_to_group = [
                val
                for val, freq in frequencies[feature].items()
                if freq < self.min_freq and val != self.str_nan
            ]
            if any(values_to_group):
                # adding default value to the order
                order.append(self.str_default)

                # grouping rare values in default value
                order.group_list(values_to_group, self.str_default)
                x_copy.loc[x_copy[feature].isin(values_to_group), feature] = self.str_default

                # updating values_orders
                self.values_orders.update({feature: order})

        # computing target rate per modality for ordering
        target_rates = x_copy.apply(target_rate, y=y, ascending=True, axis=0)

        # sorting orders based on target rates
        for feature in self.features:
            order = self.values_orders[feature]

            # new ordering according to target rate
            new_order = list(target_rates[feature])

            # leaving NaNs at the end of the list
            if self.str_nan in new_order:
                new_order.remove(self.str_nan)
                new_order += [self.str_nan]

            # sorting order updating values_orders
            self.values_orders.update({feature: order.sort_by(new_order)})

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self

class BaseQualitativeDiscretizer(GroupedListDiscretizer):
    """TODO: to implement to mutualize prepare_data and remove_feature across qualitative discretizers"""

    def __init__(
        self,
        features: List[str],
        values_orders: Dict[str, GroupedList],
        *,
        copy: bool = False,
        input_dtypes: Union[str, Dict[str, str]] = None,
        str_nan: str = None,
        verbose: bool = False
    ) -> None:
        super().__init__(features, values_orders, copy=copy, input_dtypes=input_dtypes, str_nan=str_nan, verbose=verbose)

class ChainedDiscretizer(GroupedListDiscretizer):
    """Chained Discretization based on a list of GroupedList."""

    def __init__(
        self,
        features: list[str],
        min_freq: float,
        chained_orders: list[GroupedList],
        *,
        values_orders: dict[str, GroupedList] = None,
        remove_unknown: bool = True,
        str_nan: str = "__NAN__",
        copy: bool = False,
        verbose: bool = False,
    ) -> None:
        """Initializes a ChainedDiscretizer

        Parameters
        ----------
        features : List[str]
            Columns to be bucketized
        min_freq : float
            Minimum frequency per modality
        chained_orders : List[GroupedList]
            List of modality orders
        remove_unknown : bool, optional
            Whether or not to remove unknown values. If true, they are grouped
            into the value of`str_nan` otherwise it throws an error,
            by default True
        str_nan : str, optional
            Value used to replace nan. If set, same value should be used across
            all Discretizers, by default "__NAN__"
        copy : bool, optional
            Whether or not to copy the dataset, by default False
        verbose : bool, optional
            Whether or not to print information during fit, by default False
        """
        # Initiating GroupedListDiscretizer
        super().__init__(
            features=features,
            values_orders=values_orders,
            input_dtypes='str',
            output_dtype='str',
            str_nan=str_nan,
            copy=copy,
        )

        self.min_freq = min_freq

        self.chained_orders = [GroupedList(values) for values in chained_orders]

        # parameters to handle missing/unknown values
        self.remove_unknown = remove_unknown

        # known_values: all ordered values describe in each level of the chained_orders
        # starting off with first level 
        known_values = self.chained_orders[0].values()
        # adding each level
        for next_level in self.chained_orders[1:]:
            # highest value per group of the level
            highest_ranking_value = {group: [value for value in values if value!=group][-1] for group, values in next_level.contained.items()}
            
            # adding next_level group to the order
            for group, highest_value in highest_ranking_value.items():
                highest_index = known_values.index(highest_value)
                known_values = known_values[:highest_index + 1] + [group] + known_values[highest_index + 1:]
        self.known_values = known_values

        # adding known_values to each feature's order
        for feature in self.features:
            # checking for already known values of the feature
            if feature in self.values_orders:
                order = self.values_orders[feature]
            # no known values for the feature
            else:
                order = GroupedList([])
            # checking that all values from the order are in known_values
            for value in order:
                assert value in self.known_values, f"Value {value} from feature {feature} provided in values_orders is missing from levels of chained_orders. Add value to a level of chained_orders or adapt values_orders."
            # adding known values if missing from the order
            for value in self.known_values:
                if value not in order.values():
                    order.append(value)
            order = order.sort_by(self.known_values)
            self.values_orders.update({feature: order})

        self.verbose = verbose

    def prepare_data(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Prepares the data for bucketization, checks column types.
        Converts non-string columns into strings.

        Parameters
        ----------
        X : DataFrame
            Dataset to be bucketized
        y : Series
            Model target, by default None

        Returns
        -------
        DataFrame
            Formatted X for bucketization
        """
        # copying dataframe
        x_copy = X.copy()

        # checking for ids (unique value per row)
        frequencies = x_copy[self.features].apply(
            lambda u: u.value_counts(normalize=True, dropna=False).drop(nan, errors='ignore').max(), axis=0
        )
        # for each feature, checking that at least one value is more frequent than min_freq
        for feature in self.features:
            if frequencies[feature] < self.min_freq:
                print(f"For feature '{feature}', the largest modality has {frequencies[feature]:2.2%} observations which is lower than {self.min_freq:2.2%}. This feature will not be Discretized. Consider decreasing parameter min_freq or removing this feature.")
                self.remove_feature(feature)

        # checking for columns containing floats or integers even with filled nans
        dtypes = x_copy[self.features].fillna(self.str_nan).applymap(type).apply(unique)
        not_object = dtypes.apply(lambda u: any(typ != str for typ in u))

        # non qualitative features detected
        if any(not_object):
            features_to_convert = list(not_object.index[not_object])
            if self.verbose:
                unexpected_dtypes = [typ for dtyp in dtypes[not_object] for typ in dtyp if typ != str]
                print(
                    f"""Non-string features: {str(features_to_convert)}. Trying to convert them using type_discretizers.StringDiscretizer, otherwise convert them manually. Unexpected data types: {str(list(unexpected_dtypes))}."""
                )

            # converting specified features into qualitative features
            stringer = StringDiscretizer(features=features_to_convert, values_orders=self.values_orders)
            x_copy = stringer.fit_transform(x_copy)

            # updating values_orders accordingly
            self.values_orders.update(stringer.values_orders)

        # all known values for features
        known_values = {feature: values.values() for feature, values in self.values_orders.items()}

        # checking that all unique values in X are in values_orders
        check_new_values(x_copy, self.features, known_values)

        # filling nans
        x_copy = x_copy.fillna(self.str_nan)

        return x_copy
    
    def fit(self, X: DataFrame, y: Series = None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        _type_
            _description_
        """
        # filling nans
        x_copy = self.prepare_data(X, y)

        # iterating over each feature
        for n, feature in enumerate(self.features):
            if self.verbose:  # verbose if requested
                print(f" - [ChainedDiscretizer] Fit {feature} ({n+1}/{len(self.features)})")

            # computing frequencies of each modality
            frequencies = x_copy[feature].value_counts(normalize=True)
            values, frequencies = frequencies.index, frequencies.values

            # adding NaNs to the order if any
            order = self.values_orders[feature]
            if self.str_nan in values:
                order.append(self.str_nan)

            # checking for unknown values (missing from known_values)
            missing = [value for value in values if value not in self.known_values and value != self.str_nan]

            # converting unknown values to NaN
            if self.remove_unknown & (len(missing) > 0):
                # alerting user
                print(
                    f"Order for feature '{feature}' was not provided for values:  {str(missing)}, these values will be converted to '{self.str_nan}' (policy remove_unknown=True)"
                )

                # adding missing values to the order
                order.update({self.str_nan: missing + order.get(self.str_nan)})

            # alerting user
            else:
                assert (
                    not len(missing) > 0
                ), f"Order for feature '{feature}' needs to be provided for values: {str(missing)}, otherwise set remove_unknown=True"

            # iterating over each specified orders
            for level_order in self.chained_orders:  # TODO replace all of this with labels_per_orders
                # values that are frequent enough
                to_keep = list(values[frequencies >= self.min_freq])

                # values from the order to group (not frequent enough or absent)
                values_to_group = [value for value in level_order.values() if value not in to_keep]

                # values to group into discarded values
                groups_value = [level_order.get_group(value) for value in values_to_group]

                # values of the feature to input (needed for next levels of the order)
                df_to_input = [x_copy[feature] == discarded for discarded in values_to_group]

                # inputing non frequent values
                x_copy[feature] = select(df_to_input, groups_value, default=x_copy[feature])

                # historizing in the feature's order
                for discarded, kept in zip(values_to_group, groups_value):
                    self.values_orders.get(feature).group(discarded, kept)

                # updating frequencies of each modality for the next ordering
                frequencies = x_copy[feature].value_counts(normalize=True)
                values, frequencies = frequencies.index, frequencies.values

        super().fit(X, y)

        return self


class OrdinalDiscretizer(GroupedListDiscretizer):
    """Discretizes ordered qualitative features into groups more frequent than min_freq.
    NaNs are left untouched.

    Modality is choosen amongst the preceding and following values in the provided order.
    The choosen modality is:
    - the closest in target rate
    - or the one with the lowest frequency
    """

    def __init__(
        self,
        features: list[str],
        values_orders: dict[str, GroupedList],
        min_freq: float,
        *,
        str_nan: str = "__NAN__",
        input_dtypes: Union[str, dict[str, str]] = "str",
        copy: bool = False,
        verbose: bool = False,
    ):
        """Initializes a OrdinalDiscretizer

        Parameters
        ----------
        features : List[str]
            List of column names to be discretized
        values_orders : Dict[str, Any]
            Dict of column names (keys) and modalities' associated order (values)
        min_freq : float
            Minimum frequency per modality. Less frequent modalities are grouped in the closest value of the order.
        str_nan : str, optional
            _description_, by default None
        input_dtypes : Union[str, Dict[str, str]], optional
            String of type to be considered for all features or
            Dict of column names and associated types:
            - if 'float' uses transform_quantitative
            - if 'str' uses transform_qualitative,
            default 'str'
        copy : bool, optional
            _description_, by default False
        verbose : bool, optional
            _description_, by default False
        """
        # Initiating GroupedListDiscretizer
        super().__init__(
            features=features,
            values_orders=values_orders,
            input_dtypes=input_dtypes,
            output_dtype='str',
            str_nan=str_nan,
            copy=copy,
        )

        self.min_freq = min_freq
        self.verbose = verbose

    def prepare_data(self, X: DataFrame, y: Series) -> DataFrame:
        """Called during fit step

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_

        Returns
        -------
        DataFrame
            _description_
        """
        # adding NANS
        for feature in self.features:
            if any(X[feature].isna()):
                self.values_orders[feature].append(self.str_nan)

        # removing NaNs if any already imputed
        x_copy = X
        if self.str_nan:
            x_copy = x_copy.replace(self.str_nan, nan)

        # checking for binary target
        y_values = unique(y)
        assert (0 in y_values) & (
            1 in y_values
        ), "y must be a binary Series (int or float, not object)"
        assert len(y_values) == 2, "y must be a binary Series (int or float, not object)"

        return x_copy

    def fit(self, X: DataFrame, y: Series) -> None:
        """Learns common modalities on the training set

        Parameters
        ----------
        X : DataFrame
            Training set that contains columns named after `values_orders` keys.
        y : Series
            Model target.
        """
        if self.verbose:  # verbose if requested
            print(f" - [OrdinalDiscretizer] Fit {', '.join(self.features)}")

        # checking values orders
        x_copy = self.prepare_data(X, y)

        # converting potential quantiles into there labels
        known_orders = convert_to_labels(
            features=self.features,
            quantitative_features=self.quantitative_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            dropna=True,
        )

        # grouping rare modalities for each feature
        common_modalities = x_copy[self.features].apply(
            find_common_modalities,
            y=y,
            min_freq=self.min_freq,
            values_orders=known_orders,
            axis=0,
            result_type="reduce",
        )

        # converting potential labels into there respective values (quantiles)
        self.values_orders.update(
            convert_to_values(
                features=self.features,
                quantitative_features=self.quantitative_features,
                values_orders=self.values_orders,
                label_orders=common_modalities,
                str_nan=self.str_nan,
            )
        )

        # discretizing features based on each feature's values_order
        super().fit(x_copy, y)

        return self


def find_common_modalities(
    df_feature: Series,
    y: Series,
    min_freq: float,
    values_orders: dict[str, GroupedList],
    len_df: int = None,
) -> dict[str, Any]:
    """Découpage en modalités de fréquence minimal (Cas des variables ordonnées).

    Fonction récursive : on prend l'échantillon et on cherche des valeurs sur-représentées
    Si la valeur existe on la met dans une classe et on cherche à gauche et à droite de celle-ci, d'autres variables sur-représentées
    Une fois il n'y a plus de valeurs qui soient sur-représentées,
    on fait un découpage classique avec qcut en multipliant le nombre de classes souhaité par le pourcentage de valeurs restantes sur le total

    Parameters
    ----------
    df_feature : Series
        _description_
    y : Series
        _description_
    min_freq : float
        _description_
    order : GroupedList
        _description_
    len_df : int, optional
        _description_, by default None

    Returns
    -------
    Dict[str, Any]
        _description_
    """
    # getting feature's order
    order = values_orders[df_feature.name]

    # initialisation de la taille totale du dataframe
    if len_df is None:
        len_df = len(df_feature)

    # case 1: there are missing values
    if any(isna(df_feature)):
        return find_common_modalities(
            df_feature[notna(df_feature)],
            y[notna(df_feature)],
            min_freq,
            values_orders,
            len_df,
        )

    # case 2: no missing values
    # computing frequencies and target rate of each modality
    init_frequencies = df_feature.value_counts(dropna=False, normalize=False) / len_df
    init_target_rates = y.groupby(df_feature).sum().reindex(order)

    # per-modality/value frequencies, filling missing by 0
    frequencies = init_frequencies.reindex(order, fill_value=0)

    # case 2.1: there are underrepresented modalities/values
    while any(frequencies < min_freq) & (len(frequencies) > 1):
        # updating per-group target rate per modality/value
        target_rates = series_groupy_order(init_target_rates, order) / frequencies / len_df

        # identifying the underrepresented value
        discarded_idx = argmin(frequencies)
        discarded_value = order[discarded_idx]

        # choosing amongst previous and next modality (by volume and target rate)
        kept_value = find_closest_modality(
            discarded_idx,
            order,
            list(frequencies),
            list(target_rates),
            min_freq,
        )

        # removing the value from the initial ordering
        order.group(discarded_value, kept_value)

        # removing discarded_value from frequencies
        frequencies = series_groupy_order(init_frequencies, order).fillna(0)

    # case 2.2 : no underrepresented value
    return order


def series_groupy_order(series: Series, order: GroupedList) -> Series:
    """Groups a series according to groups specified in the order

    Parameters
    ----------
    series : Series
        Values to group
    order : GroupedList
        Groups of values

    Returns
    -------
    Series
        Grouped Series
    """
    grouped_series = (
        series.groupby(list(map(order.get_group, series.index)), dropna=False).sum().reindex(order)
    )

    return grouped_series


def find_closest_modality(
    idx: int, order: GroupedList, frequencies: List[float], target_rates: Series, min_freq: float
) -> Any:
    """HELPER Finds the closest modality in terms of frequency and target rate

    Parameters
    ----------
    idx : int
        _description_
    freq_target : Series
        _description_
    min_freq : float
        _description_

    Returns
    -------
    int
        _description_
    """
    # case 1: lowest ranked modality
    if idx == 0:
        idx_closest_modality = 1

    # case 2: highest ranked modality
    elif idx == len(order) - 1:
        idx_closest_modality = len(order) - 2

    # case 3: modality ranked in the middle
    else:
        # previous modality's volume and target rate
        previous_freq, previous_target = frequencies[idx - 1], target_rates[idx - 1]

        # current modality's volume and target rate
        current_target = target_rates[idx]

        # next modality's volume and target rate
        next_freq, next_target = frequencies[idx + 1], target_rates[idx + 1]

        # identifying closest modality in terms of frequency
        least_frequent = idx - 1
        if next_freq < previous_freq:
            least_frequent = idx + 1

        # identifying closest modality in terms of target rate
        closest_target = idx - 1
        if abs(previous_target - current_target) >= abs(next_target - current_target):
            closest_target = idx + 1

        # case 3.1: grouping with the closest target rate
        idx_closest_modality = closest_target

        # case 3.2: (only) one modality isn't common, the least frequent is choosen
        if (previous_freq < min_freq < next_freq) or (next_freq < min_freq < previous_freq):
            idx_closest_modality = least_frequent

        # case 3.3: current modality is missing (no target rate), the least frequent is choosen
        if isna(current_target):
            idx_closest_modality = least_frequent

    # finding the closest value
    closest_value = order[idx_closest_modality]

    return closest_value



"""Tools for FeatureEngineering."""

from typing import Any, Dict, List

from numpy import nan, tanh
from pandas import DataFrame, Series, notna, to_datetime
from sklearn.base import BaseEstimator, TransformerMixin


class TimeDeltaConverter(BaseEstimator, TransformerMixin):
    """Converts specified DateTime columns into TimeDeltas between themselves

     - Str Columns are converted to pandas datetime columns
     - New TimeDelta column names are stored in TimeDeltaConverter.delta_features

    Parameters
    ----------
    features: List[str]
        List of DateTime columns to be converted to string.
    nans: Any, default numpy.nan
        Date value to be considered as missing data.
    drop: bool, default True
        Whether or not to drop initial DateTime columns (specified in features).
    """

    def __init__(
        self,
        features: List[str],
        nans: Any = nan,
        copy: bool = False,
        drop: bool = True,
    ) -> None:
        """_summary_

        Parameters
        ----------
        features : List[str]
            _description_
        nans : Any, optional
            _description_, by default nan
        copy : bool, optional
            _description_, by default False
        drop : bool, optional
            _description_, by default True
        """
        print("Warning: not tested for package version greater than 4.")
        self.features = features[:]
        self.copy = copy
        self.new_features: List[str] = []
        self.nans = nans
        self.drop = drop

    def fit(self, X: DataFrame, y=None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : _type_, optional
            _description_, by default None
        """
        # creating list of names of delta featuers
        for i, date1 in enumerate(self.features):
            for date2 in self.features[i + 1 :]:
                self.new_features += [f"delta_{date1}_{date2}"]

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # copying dataset
        Xc = X
        if self.copy:
            Xc = X.copy()

        # converting back nans
        if notna(self.nans):
            Xc[self.features] = Xc[self.features].where(Xc[self.features] != self.nans, nan)

        # converting to datetime
        Xc[self.features] = to_datetime(Xc[self.features].stack()).unstack()

        # iterating over each combination of dates
        for i, date1 in enumerate(self.features):
            for date2 in self.features[i + 1 :]:
                # computing timedelta of days
                Xc[f"delta_{date1}_{date2}"] = (Xc[date1] - Xc[date2]).dt.days

        # dropping date columns
        Xc = Xc.drop(self.features, axis=1)

        return Xc


class GroupNormalizer(BaseEstimator, TransformerMixin):
    """Normalizes a feature's values based on specified means and stds per groups' values

    Parameters
    ----------
    groups: List[str]
        List of qualitative features used to compute feature's mean/std per there modalities.
    features: List[str]
        List of quantitative features to be normalized.
    """

    def __init__(self, groups: List[str], features: List[str], copy: bool = True) -> None:
        """_summary_

        Parameters
        ----------
        groups : List[str]
            _description_
        features : List[str]
            _description_
        copy : bool, optional
            _description_, by default True
        """
        print("Warning: not tested for package version greater than 4.")
        self.features = features[:]
        self.groups = groups[:]

        self.copy = copy

        self.group_means = {}
        self.group_stds = {}

        self.new_features = []

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None
        """
        # iterating over each grouping column
        for group in self.groups:
            # computing group mean
            self.group_means.update({group: X.groupby(group)[self.features].mean().to_dict()})

            # computing group std
            self.group_stds.update({group: X.groupby(group)[self.features].std().to_dict()})

            # adding built features
            self.new_features += [f"{f}_norm_{group}" for f in self.features]

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # coppying dataframe
        Xc = X
        if self.copy:
            Xc = X.copy()

        # iterating over each group
        for group in self.groups:
            # computing observation level mean
            means = Xc[self.features].apply(
                lambda u: Xc[group].apply(self.group_means[group][u.name].get)
            )

            # computing observation level std
            stds = Xc[self.features].apply(
                lambda u: Xc[group].apply(self.group_stds[group][u.name].get)
            )

            # applying normalization to the feature
            Xc = Xc.join(
                Xc[self.features]
                .sub(means)
                .replace(0, nan)
                .divide(stds)
                .rename({f: f"{f}_norm_{group}" for f in self.features}, axis=1)
            )

        del means
        del stds

        return Xc


class TanhNormalizer(BaseEstimator, TransformerMixin):
    """Tanh Normalization that keeps data distribution and borns between 0 and 1

    Parameters
    ----------
    features: List[str]
        List of quantitative features to be normalized.
    """

    def __init__(self, features: List[str], copy: bool = False) -> None:
        """_summary_

        Parameters
        ----------
        features : List[str]
            _description_
        copy : bool, optional
            _description_, by default False
        """
        print("Warning: not tested for package versions greater than 4.")
        self.features = features[:]
        self.copy = copy

        self.distribs = None

    def fit(self, X: DataFrame, y=None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : _type_, optional
            _description_, by default None
        """
        self.distribs = X[self.features].agg(["mean", "std"]).to_dict(orient="index")

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # copying dataset
        Xc = X
        if self.copy:
            Xc = X.copy()

        # applying tanh normalization
        Xc[self.features] = (
            tanh(Xc[self.features].sub(self.distribs["mean"]).divide(self.distribs["std"]) * 0.01)
            + 1
        ) / 2

        return Xc


class CrossConverter(BaseEstimator, TransformerMixin):
    """Normalizes a feature's values based on specified means and stds per groups' values

    Parameters
    ----------
    features: List[str]
        List of qualitative features to be crossed should be passed through AutoCarver early on.
    """

    def __init__(self, features: List[str], copy: bool = True) -> None:
        """_summary_

        Parameters
        ----------
        features : List[str]
            _description_
        copy : bool, optional
            _description_, by default True
        """
        print("Warning: not tested for package version greater than 4.")
        self.features = features[:]
        self.copy = copy
        self.new_features: List[str] = []
        self.values: Dict[str, List[Any]] = {}

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None
        """
        # iterating over each feature
        for i, feature1 in enumerate(self.features):
            # unique values of feature1
            unq1 = X[feature1].astype(str).unique()

            for feature2 in self.features[i + 1 :]:
                # adding features names to the list of built features
                self.new_features += [f"{feature2}_x_{feature1}"]

                # unique values of feature2
                unq2 = X[feature2].astype(str).unique()

                self.values.update(
                    {f"{feature2}_x_{feature1}": [u2 + "_x_" + u1 for u2 in unq2 for u1 in unq1]}
                )

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # coppying dataframe
        Xc = X
        if self.copy:
            Xc = X.copy()

        # converting features to strings
        Xc_features = Xc[self.features].astype(str)

        # iterating over each group
        for i, feature in enumerate(self.features):
            # features to cross with
            Xc_tocross = Xc_features[self.features[i + 1 :]]

            # feature to be crossed
            Xc_crosser = Xc_features[feature]

            # crossing features
            Xc_crossed = Xc_tocross.apply(lambda u: u + "_x_" + Xc_crosser).rename(
                {f: f"{f}_x_{feature}" for f in self.features[i + 1 :]}, axis=1
            )

            # applying normalization to the feature
            Xc = Xc.join(Xc_crossed)

        del Xc_features
        del Xc_crossed
        del Xc_tocross
        del Xc_crosser

        return Xc



"""Base tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from typing import Any, Dict, List, Union

from json import dumps
from numpy import array, inf, isfinite, nan, ndarray, select, sort
from pandas import DataFrame, Series, isna, notna, unique
from sklearn.base import BaseEstimator, TransformerMixin


def nan_unique(x: Series) -> list[Any]:
    """Unique non-NaN values.

    Parameters
    ----------
    x : Series
        Values to be deduplicated.

    Returns
    -------
    List[Any]
        List of unique non-nan values
    """

    # unique values
    uniques = unique(x)

    # filtering out nans
    uniques = [u for u in uniques if notna(u)]

    return uniques


def applied_to_dict_list(applied: Union[DataFrame, Series]) -> dict[str, list[Any]]:
    """Converts a DataFrame or a List in a Dict of lists

    Parameters
    ----------
    applied : Union[DataFrame, Series]
        Result of pandas.DataFrame.apply

    Returns
    -------
    Dict[List[Any]]
        Dict of lists of rows values
    """
    # TODO: use this function whenever apply is used

    # case when it's a Series
    converted = applied.to_dict()

    # case when it's a DataFrame
    if isinstance(applied, DataFrame):
        converted = applied.to_dict(orient="list")

    return converted


# TODO: remove known_values
def check_new_values(X: DataFrame, features: List[str], known_values: Dict[str, List[Any]]) -> None:
    """Checks for new, unexpected values, in X

    Parameters
    ----------
    X : DataFrame
        New DataFrame (at transform time)
    features : List[str]
        List of column names
    known_values : Dict[str, List[Any]]
        Dict of known values per column name
    """
    # unique non-nan values in new dataframe
    uniques = X[features].apply(
        nan_unique,
        axis=0,
        result_type="expand",
    )
    uniques = applied_to_dict_list(uniques)

    # checking for unexpected values for each feature
    for feature in features:
        unexpected = [val for val in uniques[feature] if val not in known_values[feature]]
        assert (
            len(unexpected) == 0
        ), f"Unexpected value! The ordering for values: {str(list(unexpected))} of feature '{feature}' was not provided. There might be new values in your test/dev set. Consider taking a bigger test/dev set or dropping the column {feature}."


def check_missing_values(
    X: DataFrame, features: List[str], known_values: Dict[str, List[Any]]
) -> None:
    """Checks for missing values from X, (unexpected values in values_orders)

    Parameters
    ----------
    X : DataFrame
        New DataFrame (at transform time)
    features : List[str]
        List of column names
    known_values : Dict[str, List[Any]]
        Dict of known values per column name
    """
    # unique non-nan values in new dataframe
    uniques = X[features].apply(
        nan_unique,
        axis=0,
        result_type="expand",
    )
    uniques = applied_to_dict_list(uniques)

    # checking for unexpected values for each feature
    for feature in features:
        unexpected = [val for val in known_values[feature] if val not in uniques[feature]]
        assert (
            len(unexpected) == 0
        ), f"Unexpected value! The ordering for values: {str(list(unexpected))} of feature '{feature}' was provided but there are not matching value in provided X. You should check 'values_orders['{feature}']' for unwanted values."


class GroupedList(list):
    """An ordered list that's extended with a dict."""

    def __init__(self, iterable: Any = ()) -> None:
        """An ordered list that historizes its elements' merges.

        Parameters
        ----------
        iterable : Any, optional
            list, dict or GroupedList, by default ()
        """

        # case -1: iterable is an array
        if isinstance(iterable, ndarray):
            iterable = list(iterable)

        # case 0: iterable is the contained dict
        if isinstance(iterable, dict):
            # storing ordered keys of the dict
            keys = list(iterable)

            # storing the values contained per key
            self.contained = dict(iterable.items())

            # checking that all values are only present once
            all_values = [val for _, values in iterable.items() for val in values]
            assert len(list(set(all_values))) == len(
                all_values
            ), "A value is present in several keys (groups)"

            # adding key to itself if it's not present in an other key
            keys_copy = keys[:]  # copying initial keys
            for key in keys_copy:
                # checking that the value is not comprised in an other key
                all_values = [
                    val
                    for iter_key, values in iterable.items()
                    for val in values
                    if key != iter_key
                ]
                if key not in all_values:
                    # checking that key is missing from its values
                    if key not in iterable[key]:
                        self.contained.update({key: self.contained[key] + [key]})
                # the key already is in another key (and its values are empty)
                # the key as already been grouped
                else:
                    self.contained.pop(key)
                    keys.remove(key)

            # initiating the list with those keys
            super().__init__(keys)

        # case 1: copying a GroupedList
        elif hasattr(iterable, "contained"):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # copying values associated to keys
            self.contained = dict(iterable.contained.items())

        # case 2: initiating GroupedList from a list
        elif isinstance(iterable, list):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # initiating the values with the provided list of keys
            self.contained = {v: [v] for v in iterable}

    def get(self, key: Any) -> List[Any]:
        """List of values contained in key

        Parameters
        ----------
        key : Any
            Group.

        Returns
        -------
        List[Any]
            Values contained in key
        """

        # default to fing an element
        found = self.contained.get(key)

        return found

    def group(self, discarded: Any, kept: Any) -> None:
        """Groups the discarded value with the kept value

        Parameters
        ----------
        discarded : Any
            Value to be grouped into the key `to_keep`.
        kept : Any
            Key value in which to group `discarded`.
        """

        # checking that those values are distinct
        if not is_equal(discarded, kept):
            # checking that those values exist in the list
            assert discarded in self, f"{discarded} not in list"
            assert kept in self, f"{kept} not in list"

            # accessing values contained in each value
            contained_discarded = self.contained.get(discarded)
            contained_kept = self.contained.get(kept)

            # updating contained dict
            self.contained.update({kept: contained_discarded + contained_kept, discarded: []})

            # removing discarded from the list
            self.remove(discarded)

    def group_list(self, to_discard: List[Any], to_keep: Any) -> None:
        """Groups elements to_discard into values to_keep

        Parameters
        ----------
        to_discard : List[Any]
            Values to be grouped into the key `to_keep`.
        to_keep : Any
            Key value in which to group `to_discard` values.
        """

        for discarded, kept in zip(to_discard, [to_keep] * len(to_discard)):
            self.group(discarded, kept)

    def append(self, new_value: Any) -> None:
        """Appends a new_value to the GroupedList

        Parameters
        ----------
        new_value : Any
            New key to be added.
        """

        self += [new_value]
        self.contained.update({new_value: [new_value]})

    def update(self, new_value: Dict[Any, List[Any]]) -> None:
        """Updates the GroupedList via a dict

        Parameters
        ----------
        new_value : Dict[Any, List[Any]]
            Dict of key, values to updated `contained` dict
        """

        # adding keys to the order if they are new values
        self += [key for key, _ in new_value.items() if key not in self]

        # updating contained according to new_value
        self.contained.update(new_value)

    def sort(self):
        """Sorts the values of the list and dict (if any, NaNs are last).

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """
        # str values
        keys_str = [key for key in self if isinstance(key, str)]

        # non-str values
        keys_float = [key for key in self if not isinstance(key, str)]

        # sorting and merging keys
        keys = list(sort(keys_str)) + list(sort(keys_float))

        # recreating an ordered GroupedList
        sorted = GroupedList({k: self.get(k) for k in keys})

        return sorted

    def sort_by(self, ordering: List[Any]) -> None:
        """Sorts the values of the list and dict according to `ordering`, if any, NaNs are the last.

        Parameters
        ----------
        ordering : List[Any]
            Order used for ordering of the list of keys.

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """

        # checking that all values are given an order
        assert all(
            o in self for o in ordering
        ), f"Unknown values in ordering: {str([v for v in ordering if v not in self])}"
        assert all(
            s in ordering for s in self
        ), f"Missing value from ordering: {str([v for v in self if v not in ordering])}"

        # ordering the contained
        sorted = GroupedList({k: self.get(k) for k in ordering})

        return sorted

    def remove(self, value: Any) -> None:
        """Removes a value from the GroupedList

        Parameters
        ----------
        value : Any
            value to be removed
        """
        super().remove(value)
        self.contained.pop(value)

    def pop(self, idx: int) -> None:
        """Pop a value from the GroupedList by index

        Parameters
        ----------
        idx : int
            Index of the value to be popped out
        """
        value = self[idx]
        self.remove(value)

    def get_group(self, value: Any) -> Any:
        """Returns the key (group) containing the specified value

        Parameters
        ----------
        value : Any
            Value for which to find the group.

        Returns
        -------
        Any
            Corresponding key (group)
        """

        found = [
            key
            for key, values in self.contained.items()
            if any(is_equal(value, elt) for elt in values)
        ]

        if any(found):
            return found[0]

        return value

    def values(self) -> List[Any]:
        """All values contained in all groups

        Returns
        -------
        List[Any]
            List of all values in the GroupedList
        """

        known = [value for values in self.contained.values() for value in values]

        return known

    def contains(self, value: Any) -> bool:
        """Checks if a value is contained in any group, also matches NaNs.

        Parameters
        ----------
        value : Any
            Value to search for

        Returns
        -------
        bool
            Whether the value is in the GroupedList
        """

        known_values = self.values()

        return any(is_equal(value, known) for known in known_values)

    def get_repr(self, char_limit: int = 6) -> List[str]:
        """Returns a representative list of strings of values of groups.

        Parameters
        ----------
        char_limit : int, optional
            Maximum number of character per string, by default 6

        Returns
        -------
        List[str]
            List of short str representation of the keys' values
        """

        # initiating list of group representation
        repr: List[str] = []

        # iterating over each group
        for group in self:
            # accessing group's values
            values = self.get(group)

            if len(values) == 0:  # case 0: there are no value in this group
                pass

            elif len(values) == 1:  # case 1: there is only one value in this group
                repr += [values[0]]

            elif len(values) == 2:  # case 2: two values in this group
                repr += [f"{values[1]}"[:char_limit] + " and " + f"{values[0]}"[:char_limit]]

            elif len(values) > 2:  # case 3: more than two values in this group
                repr += [f"{values[-1]}"[:char_limit] + " to " + f"{values[0]}"[:char_limit]]

        return repr


# TODO: add a summary
# TODO: output a json
# TODO: add a base discretizer that implements prepare_data (add reset_index ? -> add duplicate column check)
class GroupedListDiscretizer(BaseEstimator, TransformerMixin):
    """Discretizer that uses a dict of grouped values."""

    def __init__(
        self,
        features: list[str],
        *,
        values_orders: dict[str, GroupedList] = None,
        input_dtypes: Union[str, dict[str, str]] = 'str',
        output_dtype: str = 'str',
        str_nan: str = None,
        copy: bool = False,
    ) -> None:
        """Initiates a Discretizer by dict of GroupedList

        Parameters
        ----------
        features : list[str]
            List of column names to be discretized
        values_orders : dict[str, GroupedList], optional
            Per feature ordering, by default None
        input_dtypes : Union[str, dict[str, str]], optional
            String of type to be considered for all features or
            Dict of column names and associated types:
            - if 'float' uses transform_quantitative
            - if 'str' uses transform_qualitative,
            by default 'str'
        output_dtype : str, optional
            _description_, by default 'str'
        str_nan : str, optional
            Default string value attributed to nan, by default None
        copy : bool, optional
            Whether or not to copy the input DataFrame, by default False
        """
        self.features = list(set(features))
        if values_orders is None:
            values_orders = {}
        self.values_orders = {feature: GroupedList(values) for feature, values in values_orders.items()}
        self.copy = copy

        # input feature types
        if isinstance(input_dtypes, str):
            input_dtypes = {feature: input_dtypes for feature in features}
        self.input_dtypes = input_dtypes

        # output type
        self.output_dtype = output_dtype

        # string value of numpy.nan
        self.str_nan = str_nan

        # identifying qualitative features by there type
        self.qualitative_features = [
            feature for feature in features if self.input_dtypes[feature] == "str"
        ]

        # identifying quantitative features by there type
        self.quantitative_features = [
            feature for feature in features if self.input_dtypes[feature] == "float"
        ]

        # for each feature, getting label associated to each value
        self.labels_per_values = {}  # will be initiated during fit

    def get_labels_per_values(self) -> dict[str, dict[Any, Any]]:
        """Creates a dict that contains, for each feature, for each value, the associated label

        Returns
        -------
        dict[str, dict[Any, Any]]
            Dict of labels per values per feature
        """
        # initiating dict of labels per values per feature
        labels_per_values = {}

        # iterating over each feature
        for feature in self.features:
            # known values of the feature(grouped)
            values = self.values_orders[feature]

            # case 0: quantitative feature -> labels per quantile (removes str_nan)
            if feature in self.quantitative_features:
                labels = get_labels(values, self.str_nan)
            # case 1: qualitative feature -> by default, labels are values
            else:
                labels = values[:]
            
            # case 2: requested float output -> converting to integers
            if self.output_dtype == 'float':
                labels = [n for n, _ in enumerate(labels)]

            # building label per value
            label_per_value = {}
            for group_of_values, label in zip(values, labels):
                for value in values.get(group_of_values):
                    label_per_value.update({value: label})
            
            # storing in full dict
            labels_per_values.update({feature: label_per_value})
        
        return labels_per_values

    def remove_feature(self, feature: str) -> None:
        """Removes a feature from the Discretizer

        Parameters
        ----------
        feature : str
            Column name of the feature
        """
        if feature in self.features:
            self.features.remove(feature)
            if feature in self.qualitative_features:
                self.qualitative_features.remove(feature)
            if feature in self.quantitative_features:
                self.quantitative_features.remove(feature)
            if feature in self.values_orders:
                self.values_orders.pop(feature)
            if feature in self.input_dtypes:
                self.input_dtypes.pop(feature)
    
    def prepare_data(self, X: DataFrame, y: Series = None) -> DataFrame:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series, optional
            _description_, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # copying X
        x_copy = X
        if self.copy and X is not None:
            missing_columns = [feature for feature in self.features if feature not in X]
            assert len(missing_columns) == 0, f"Missing features from the provided DataFrame: {str(missing_columns)}"
            x_copy = X.copy()

        # checking for binary target
        if y is not None:
            y_values = unique(y)
            assert (0 in y_values) & (
                1 in y_values
            ), "y must be a binary Series (int or float, not object)"
            assert len(y_values) == 2, "y must be a binary Series (int or float, not object)"

        return x_copy

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """Learns the labels associated to each value for each feature

        Parameters
        ----------
        X : DataFrame
            Contains columns named after `features` attribute
        y : Series, optional
            Model target, by default None
        """
        # checking that all features to discretize are in values_orders
        missing_features = [feature for feature in self.features if feature not in self.values_orders]
        assert len(missing_features) == 0, f"Missing values_orders for following features {str(missing_features)}."

        # for each feature, getting label associated to each value
        self.labels_per_values = self.get_labels_per_values()

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Groups values of features (values_orders keys) according to
        there corresponding GroupedList (values_orders values) based on
        the `GroupedList.contained` dict.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after `features` attribute
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            _description_
        """

        # copying dataframes
        x_copy = X
        if self.copy:
            x_copy = X.copy()

        # transforming quantitative features
        if len(self.quantitative_features) > 0:
            x_copy = self.transform_quantitative(x_copy, y)

        # transforming qualitative features
        if len(self.qualitative_features) > 0:
            x_copy = self.transform_qualitative(x_copy, y)

        return x_copy

    def transform_quantitative(self, X: DataFrame, y: Series) -> DataFrame:
        """Groups values of features (values_orders keys) according to
        there corresponding GroupedList (values_orders values) based on
        the `GroupedList.contained` dict.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after `values_orders` keys
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # dataset length
        x_len = X.shape[0]

        # grouping each quantitative feature
        for feature in self.quantitative_features:
            # feature's labels associated to each quantile
            feature_values = self.values_orders[feature]

            # keeping track of nans
            nans = isna(X[feature])

            # converting nans to there corresponding quantile (if it was grouped to a quantile)
            if any(nans):
                assert (
                     feature_values.contains(self.str_nan)
                ), f"Unexpected value! Missing values found for feature '{feature}' at transform step but not during fit. There might be new values in your test/dev set. Consider taking a bigger test/dev set or dropping the column {feature}."
                nan_value = feature_values.get_group(self.str_nan)
                # checking that nans have been grouped to a quantile otherwise they are left as numpy.nan (for comparison purposes)
                if nan_value != self.str_nan:
                    X.loc[nans, feature] = nan_value

            # removing nans from quantiles and labels (can not mix str and floats for comparison purposes)
            if self.str_nan in feature_values:  # filtering out nans if any
                feature_values = [value for value in feature_values if value != self.str_nan]

            # list of masks of values to replace with there respective group
            values_to_group = [X[feature] <= quantile for quantile in feature_values]

            # corressponding group for each value
            group_labels = [[self.labels_per_values[feature][value]] * x_len for value in feature_values]

            # checking for values to group
            if len(values_to_group) > 0:
                X[feature] = select(values_to_group, group_labels, default=X[feature])

            # converting nans to there value
            if any(nans):
                X.loc[nans, feature] = self.labels_per_values[feature].get(nan_value, self.str_nan)

        return X

    def transform_qualitative(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Groups values of features (values_orders keys) according to
        there corresponding GroupedList (values_orders values) based on
        the `GroupedList.contained` dict.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after `values_orders` keys
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            _description_
        """
        # filling up nans with specified value
        if self.str_nan:
            X[self.qualitative_features] = X[self.qualitative_features].fillna(self.str_nan)

        # checking that all unique values in X are in values_orders
        check_new_values(X, self.qualitative_features, {feature: values.values() for feature, values in self.values_orders.items()})

        # replacing values for there corresponding label
        X = X.replace({feature: label_per_value for feature, label_per_value in self.labels_per_values.items() if feature in self.qualitative_features})           

        return X
    
    def to_json(self) ->  str:
        """Converts the GroupedListDiscretizer's values_orders to .json

        Returns
        -------
        str
            _description_
        """
        # extracting contained dictionnaries
        content = {
            "features": self.features,
            "values_ordres": self.values_orders,  # TODO: adapt maybe init GroupedList with {"values": values, "contained": values.contained} ?
            "input_dtypes": self.input_dtypes,
            "output_dtype": self.output_dtype,
            "str_nan": self.str_nan,
            "copy": self.copy,
        }
        # contained_orders = {feature: {"values": values, "contained": values.contained} for feature, values in self.values_orders.items()}
        # dumping as json
        return dumps(content)
    
    def summary(self) -> DataFrame:
        """Summarizes the data bucketization

        Returns
        -------
        DataFrame
            A summary of feature's values
        """
        # getting quantitative labels
        if any(self.quantitative_features):
            print("converting to labels\n\n")
            labels_orders = convert_to_labels(
                self.features, self.quantitative_features, self.values_orders, self.str_nan, dropna=False
            )
        # initiating summaries
        summaries: list[dict[str, Any]] = []
        for feature in self.features:
            init_summary = {'feature': feature}  # initiating feature summary
            
            # case 0: quantitative features
            if feature in self.quantitative_features:
                init_summary.update({'data_type': 'quantitative'})
                # feature's modalities
                modalities = labels_orders[feature]
                # values included in each modality
                contained = modalities.contained
                
                # adding nans
                if self.values_orders[feature].contains(self.str_nan):
                    nan_group = self.values_orders[feature].get_group(self.str_nan)
                    contained.update({nan_group: contained[nan_group] + [self.str_nan]})


            # case 1: quantitative features
            else:
                init_summary.update({'data_type': 'qualitative'})
                # feature's modalities
                modalities = self.values_orders[feature]
                # values included in each modality
                contained = modalities.contained
            
            # adding all modalities and there content
            for modality in modalities:
                modality_summary = {k: v for k, v in init_summary.items()}
                modality_summary.update({"modality": modality, "values": contained[modality]})

                # case 2: when the output_dtype == 'float'
                #if self.output_dtype == 'float':
                #    modality_summary.update({"modality": self.output_dtype[feature][modality]})

                summaries += [modality_summary]

        return summaries


def convert_to_labels(
    features: List[str],
    quantitative_features: List[str],
    values_orders: Dict[str, GroupedList],
    str_nan: str,
    dropna: bool = True,
) -> Dict[str, GroupedList]:
    """Converts a values_orders values to labels (quantiles)"""

    # copying values_orders without nans
    labels_orders = {
        feature: GroupedList([value for value in values_orders[feature] if value != str_nan])
        for feature in features
    }

    # for quantitative features getting labels per quantile
    if any(quantitative_features):
        # getting group "name" per quantile
        quantiles_labels = get_quantiles_labels(quantitative_features, values_orders, str_nan)

        # applying alliases to known orders
        for feature in quantitative_features:
            labels_orders.update(
                {
                    feature: GroupedList(
                        [quantiles_labels[feature][quantile] for quantile in labels_orders[feature]]
                    )
                }
            )

    # adding back nans if requested
    if not dropna:
        for feature in features:
            if str_nan in values_orders[feature]:
                order = labels_orders[feature]
                order.append(str_nan)  # adding back nans at the end of the order
                labels_orders.update({feature: order})



    return labels_orders


def convert_to_values(
    features: list[str],
    quantitative_features: list[str],
    values_orders: dict[str, GroupedList],
    label_orders: dict[str, GroupedList],
    str_nan: str,
):
    # for quantitative features getting labels per quantile
    if any(quantitative_features):
        # getting quantile per group "name"
        labels_to_quantiles = get_labels_quantiles(quantitative_features, values_orders, str_nan)

    # updating feature orders (that keeps NaNs and quantiles)
    for feature in features:
        # initial complete ordering with NAN and quantiles
        order = values_orders[feature]

        # checking for grouped modalities
        groups_to_discard = label_orders[feature].contained

        # grouping the raw quantile values
        for kept_value, group_to_discard in groups_to_discard.items():
            # for qualitative features grouping as is
            # for quantitative features getting quantile per alias
            if feature in quantitative_features:
                # getting raw quantiles to be grouped
                group_to_discard = [
                    labels_to_quantiles[feature][label_discarded]
                    if label_discarded != str_nan
                    else str_nan
                    for label_discarded in group_to_discard
                ]

                # choosing the value to keep as the group
                which_to_keep = [value for value in group_to_discard if value != str_nan]
                # case 0: keeping the largest value amongst the discarded (otherwise they wont be grouped)
                if len(which_to_keep) > 0:
                    kept_value = max(which_to_keep)
                # case 1: there is only str_nan in the group (it was not grouped)
                else:
                    kept_value = group_to_discard[0]

            # grouping quantiles
            order.group_list(group_to_discard, kept_value)

        # updating ordering
        values_orders.update({feature: order})

    return values_orders


def min_value_counts(
    x: Series,
    values_orders: Dict[str, GroupedList] = None,
    dropna: bool = False,
    normalize: bool = True,
) -> float:
    """Minimum of modalities' frequencies.

    Parameters
    ----------
    x : Series
        _description_
    values_orders : Dict[str, GroupedList]
        _description_

    Returns
    -------
    float
        _description_
    """
    # modality frequency
    values = x.value_counts(dropna=dropna, normalize=normalize)

    # setting indices
    order = values_orders.get(x.name)
    if order is not None:
        values = values.reindex(order).fillna(0)

    # minimal frequency
    min_values = values.values.min()

    return min_values


def value_counts(x: Series, dropna: bool = False, normalize: bool = True) -> dict:
    """Counts the values of each modality of a series into a dictionnary

    Parameters
    ----------
    x : Series
        _description_
    dropna : bool, optional
        _description_, by default False
    normalize : bool, optional
        _description_, by default True

    Returns
    -------
    dict
        _description_
    """

    values = x.value_counts(dropna=dropna, normalize=normalize)

    return values.to_dict()


def target_rate(x: Series, y: Series, dropna: bool = True, ascending=True) -> dict:
    """Target y rate per modality of x into a dictionnary

    Parameters
    ----------
    x : Series
        _description_
    y : Series
        _description_
    dropna : bool, optional
        _description_, by default True
    ascending : bool, optional
        _description_, by default True

    Returns
    -------
    dict
        _description_
    """

    rates = y.groupby(x, dropna=dropna).mean().sort_values(ascending=ascending)

    return rates.to_dict()


def nunique(x: Series, dropna=False) -> int:
    """Computes number of unique modalities

    Parameters
    ----------
    x : Series
        _description_
    dropna : bool, optional
        _description_, by default False

    Returns
    -------
    int
        _description_
    """

    uniques = unique(x)
    n = len(uniques)

    # removing nans
    if dropna:
        if any(isna(uniques)):
            n -= 1

    return n


def is_equal(a: Any, b: Any) -> bool:
    """Checks if a and b are equal (NaN insensitive)

    Parameters
    ----------
    a : Any
        _description_
    b : Any
        _description_

    Returns
    -------
    bool
        _description_
    """

    # default equality
    equal = a == b

    # Case where a and b are NaNs
    if isna(a) and isna(b):
        equal = True

    return equal


def get_labels(quantiles: List[float], str_nan: str) -> List[str]:
    """_summary_

    Parameters
    ----------
    feature : str
        _description_
    order : GroupedList
        _description_
    str_nan : str
        _description_

    Returns
    -------
    List[str]
        _description_
    """
    # filtering out nan for formatting
    if str_nan in quantiles:
        quantiles = [val for val in quantiles if val != str_nan]

    # filtering out inf for formatting
    if inf in quantiles:
        quantiles = [val for val in quantiles if isfinite(val)]

    # converting quantiles in string
    labels = format_quantiles(quantiles)

    return labels


def get_quantiles_labels(
    features: List[str], values_orders: Dict[str, GroupedList], str_nan: str
) -> Dict[str, GroupedList]:
    """Converts a values_orders of quantiles into a values_orders of string quantiles

    Parameters
    ----------
    features : List[str]
        _description_
    values_orders : Dict[str, GroupedList]
        _description_
    str_nan : str
        _description_

    Returns
    -------
    Dict[str, GroupedList]
        _description_
    """
    # applying quantiles formatting to orders of specified features
    quantiles_to_labels = {}
    for feature in features:
        quantiles = list(values_orders[feature])
        labels = get_labels(quantiles, str_nan)
        # associates quantiles to their respective labels
        quantiles_to_labels.update(
            {feature: {quantile: alias for quantile, alias in zip(quantiles, labels)}}
        )

    return quantiles_to_labels


def get_labels_quantiles(
    features: List[str], values_orders: Dict[str, GroupedList], str_nan: str
) -> Dict[str, GroupedList]:
    """Converts a values_orders of quantiles into a values_orders of string quantiles

    Parameters
    ----------
    features : List[str]
        _description_
    values_orders : Dict[str, GroupedList]
        _description_
    str_nan : str
        _description_

    Returns
    -------
    Dict[str, GroupedList]
        _description_
    """
    # applying quantiles formatting to orders of specified features
    labels_to_quantiles = {}
    for feature in features:
        quantiles = list(values_orders[feature])
        labels = get_labels(quantiles, str_nan)
        # associates quantiles to their respective labels
        labels_to_quantiles.update(
            {feature: {alias: quantile for quantile, alias in zip(quantiles, labels)}}
        )

    return labels_to_quantiles


def format_quantiles(a_list: List[float]) -> List[str]:
    """Formats a list of float quantiles into a list of boundaries.

    Rounds quantiles to the closest power of 1000.

    Parameters
    ----------
    a_list : List[float]
        Sorted list of quantiles to convert into string

    Returns
    -------
    List[str]
        List of boundaries per quantile
    """

    # finding the closest power of thousands for each element
    closest_powers = [
        next((k for k in range(-3, 4) if abs(elt) / 100 // 1000 ** (k) < 10)) for elt in a_list
    ]

    # rounding elements to the closest power of thousands
    rounded_to_powers = [elt / 1000 ** (k) for elt, k in zip(a_list, closest_powers)]

    # computing optimal decimal per unique power of thousands
    optimal_decimals: Dict[str, int] = {}
    for power in unique(closest_powers):  # iterating over each power of thousands found
        # filtering on the specific power of thousands
        sub_array = array([elt for elt, p in zip(rounded_to_powers, closest_powers) if power == p])

        # number of distinct values
        n_uniques = sub_array.shape[0]

        # computing the first rounding decimal that allows for distinction of
        # each values when rounded, by default None (no rounding)
        optimal_decimal = next(
            (k for k in range(1, 10) if len(unique(sub_array.round(k))) == n_uniques),
            None,
        )

        # storing in the dict
        optimal_decimals.update({power: optimal_decimal})

    # rounding according to each optimal_decimal
    rounded_list: List[float] = []
    for elt, power in zip(rounded_to_powers, closest_powers):
        # rounding each element
        rounded = elt  # by default None (no rounding)
        optimal_decimal = optimal_decimals.get(power)
        if optimal_decimal:  # checking that the optimal decimal exists
            rounded = round(elt, optimal_decimal)

        # adding the rounded number to the list
        rounded_list += [rounded]

    # dict of suffixes per power of thousands
    suffixes = {-3: "n", -2: "mi", -1: "m", 0: "", 1: "K", 2: "M", 3: "B"}

    # adding the suffixes
    formatted_list = [
        f"{elt: 1.1f}{suffixes[power]}" for elt, power in zip(rounded_list, closest_powers)
    ]

    # keeping zeros
    formatted_list = [
        rounded if raw != 0 else f"{raw: 1.1f}" for rounded, raw in zip(formatted_list, a_list)
    ]

    # stripping whitespaces
    formatted_list = [string.strip() for string in formatted_list]

    # low and high bounds per quantiles
    upper_bounds = formatted_list + [nan]
    lower_bounds = [nan] + formatted_list
    order: List[str] = []
    for lower, upper in zip(lower_bounds, upper_bounds):
        if isna(lower):
            order += [f"x <= {upper}"]
        elif isna(upper):
            order += [f"{lower} < x"]
        else:
            order += [f"{lower} < x <= {upper}"]

    return order



"""Set of tests for base_discretizers module."""

from AutoCarver.discretizers.utils.base_discretizers import *
from pytest import raises
from numpy import nan

def test_grouped_list_init():
    """ Tests the initialization of a GroupedList"""
    
    # init by list
    test_list = ['1', '2', '3']
    groupedlist = GroupedList(test_list)
    assert groupedlist == test_list, "When init by list, GroupedList.contained should have only one value per key"
    assert groupedlist.contained == {'1': ['1'], '2': ['2'], '3': ['3']}, "When init by list, GroupedList.contained should have only one value per key: itself"
    
    # init by dict
    test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3']}
    groupedlist = GroupedList(test_dict)
    assert groupedlist == ['1', '2', '3'], "When init by dict, all keys should be in the list"
    assert groupedlist.contained == {'1': ['1', '4'], '2': ['2'], '3': ['3']},"When init by dict, all values should be in stored in the contained dict"

    test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3'], '4': []}
    groupedlist = GroupedList(test_dict)
    assert groupedlist == ['1', '2', '3'], "When init by dict, keys that are in another key (group) should be popped"
    assert groupedlist.contained == {'1': ['1', '4'], '2': ['2'], '3': ['3']},"When init by dict, all values should be in stored in the contained dict"

    test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3'], '4': []}
    groupedlist = GroupedList(test_dict)
    assert groupedlist == ['1', '2', '3'], "When init by dict, keys that are in another key (group) should be popped"
    assert groupedlist.contained == {'1': ['1', '4'], '2': ['2'], '3': ['3']},"When init by dict, all values should be in stored in the contained dict"

    # check that a value can not be in several keys (groups)
    with raises(AssertionError):
        test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3'], '4': ['4']}
        groupedlist = GroupedList(test_dict)
    
    test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3'], '4': [], '5': []}
    groupedlist = GroupedList(test_dict)
    assert groupedlist == ['1', '2', '3', '5'], "When init by dict, keys that are in no key (group) should be kept in the list"
    assert groupedlist.contained == {'1': ['1', '4'], '2': ['2'], '3': ['3'], '5': ['5']},"When init by dict, keys that are in no key (group) should be added to themselves"
    
    # init by copy
    groupedlist_copy = GroupedList(groupedlist)
    assert groupedlist_copy == ['1', '2', '3', '5'], "When init by GroupedList, GroupedList should be an exact copy"
    assert groupedlist_copy.contained == {'1': ['1', '4'], '2': ['2'], '3': ['3'], '5': ['5']}, "When init by GroupedList, GroupedList should be an exact copy"


def test_grouped_list_functions():
    """Tests GroupedList functions"""
    
    # init a groupedlist
    test_dict = {'1': ['1', '4'], '2': ['2'], '3': ['3'], '4': [], '5': []}
    groupedlist = GroupedList(test_dict)
    
    # test get
    assert groupedlist.get('1') == ['1', '4']
    
    
    # test group
    groupedlist.group('1', '5')
    assert groupedlist == ['2', '3', '5']
    assert groupedlist.contained == {'2': ['2'], '3': ['3'], '5': [ '1', '4', '5']}
    with raises(AssertionError):
        groupedlist.group('7', '5')
        groupedlist.group('5', '8')
    
    # test group_list
    groupedlist.group_list(['2', '3'], '5')
    assert groupedlist == ['5']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5']}
    
    # test append
    groupedlist.append('0')
    assert groupedlist == ['5', '0']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5'], '0': ['0']}
    
    # test update
    groupedlist.update({'0': ['0', '7']})
    assert groupedlist == ['5', '0']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5'], '0': ['0', '7']}
    
    # test sort
    groupedlist = groupedlist.sort()
    assert groupedlist == ['0', '5']
    assert groupedlist.contained == {'0': ['0', '7'], '5': [ '3', '2', '1', '4', '5']}
    
    # test sort_by
    groupedlist = groupedlist.sort_by(['5', '0'])
    assert groupedlist == ['5', '0']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5'], '0': ['0', '7']}
    with raises(AssertionError):
        groupedlist.sort_by(['5', '0', '1'])
        groupedlist.sort_by(['5'])
    
    # test remove
    groupedlist.append('15')
    groupedlist.remove('15')
    assert groupedlist == ['5', '0']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5'], '0': ['0', '7']}
    
    # test pop
    groupedlist.append('15')
    groupedlist.pop(len(groupedlist) - 1)
    assert groupedlist == ['5', '0']
    assert groupedlist.contained == {'5': [ '3', '2', '1', '4', '5'], '0': ['0', '7']}
    
    # test get_group
    groupedlist.append('15')
    groupedlist.pop(len(groupedlist) - 1)
    assert groupedlist.get_group('3') == '5'
    assert groupedlist.get_group('0') == '0'
    
    # test values
    assert groupedlist.values() == [ '3', '2', '1', '4', '5', '0', '7']
    
    # test contains
    assert groupedlist.contains('3')
    assert not groupedlist.contains('15')
    
    # test get_repr
    assert groupedlist.get_repr() == ['5 to 3', '7 and 0']

# TODO: test quantitative discretization
def test_grouped_list_discretizer(x_train: DataFrame, x_test_1: DataFrame, x_test_2: DataFrame):
    """Tests GroupedListDiscretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    x_test_1 : DataFrame
        Simulated Test DataFrame
    x_test_2 : DataFrame
        Simulated Test DataFrame
    """
    
    # values to input nans
    str_nan = '__NAN__'
    
    # defining values_orders
    order = ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+']
    
    # ordering for base qualitative ordinal feature
    groupedlist = GroupedList(order)
    groupedlist.group_list(['Low-', 'Low'], 'Low+')
    groupedlist.group_list(['Medium+', 'High-'], 'High')

    # ordering for qualitative ordinal feature that contains NaNs
    groupedlist_lownan = GroupedList(order + [str_nan])
    groupedlist_lownan.group_list(['Low-', 'Low'], 'Low+')
    groupedlist_lownan.group_list(['Medium+', 'High-'], 'High')

    # storing per feature orders
    values_orders = {
        "Qualitative_Ordinal": groupedlist,
        "Qualitative_Ordinal_lownan": groupedlist_lownan,
    }
    features = ["Qualitative_Ordinal", "Qualitative_Ordinal_lownan"]
    
    # initiating discretizer 
    discretizer = GroupedListDiscretizer(features=features, values_orders=values_orders, str_nan=str_nan, input_dtypes='str', copy=True)
    x_discretized = discretizer.fit_transform(x_train)

    # testing ordinal qualitative feature discretization
    x_expected = x_train.copy()
    x_expected["Qualitative_Ordinal"] = x_expected["Qualitative_Ordinal"].replace('Low-', 'Low+')\
        .replace('Low', 'Low+')\
        .replace('Low+', 'Low+')\
        .replace('Medium-', 'Medium-')\
        .replace('Medium', 'Medium')\
        .replace('Medium+', 'High')\
        .replace('High-', 'High')\
        .replace('High', 'High')\
        .replace('High+', 'High+')
    assert all(x_expected["Qualitative_Ordinal"] == x_discretized["Qualitative_Ordinal"]), "incorrect discretization"

    # testing ordinal qualitative feature discretization with nans
    x_expected = x_train.copy()
    x_expected["Qualitative_Ordinal_lownan"] = x_expected["Qualitative_Ordinal_lownan"].replace('Low-', 'Low+')\
        .replace('Low', 'Low+')\
        .replace('Low+', 'Low+')\
        .replace('Medium-', 'Medium-')\
        .replace('Medium', 'Medium')\
        .replace('Medium+', 'High')\
        .replace('High-', 'High')\
        .replace('High', 'High')\
        .replace('High+', 'High+')\
        .replace(nan, '__NAN__')
    assert all(x_expected["Qualitative_Ordinal_lownan"] == x_discretized["Qualitative_Ordinal_lownan"]), "incorrect discretization with nans"

    # checking that other columns are left unchanged
    assert all(x_discretized["Quantitative"] == x_discretized["Quantitative"]), "Other columns should not be modified"



"""Set of tests for quantitative_discretizers module."""

from AutoCarver.discretizers.utils.type_discretizers import *
from pandas import DataFrame

def test_string_discretizer(x_train: DataFrame) -> None:
    """Tests StringDiscretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    """

    qualitative_features = ["Qualitative", "Qualitative_grouped", "Qualitative_lownan", "Qualitative_highnan", "Discrete_Qualitative_noorder", "Discrete_Qualitative_lownan_noorder", "Discrete_Qualitative_rarevalue_noorder"]
    ordinal_features = ["Qualitative_Ordinal_lownan", "Discrete_Qualitative_highnan"]
    values_orders = {
        "Qualitative_Ordinal": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Qualitative_Ordinal_lownan": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Discrete_Qualitative_highnan" : ["1", "2", "3", "4", "5", "6", "7"],
    }


    discretizer = StringDiscretizer(
        features=list(set(ordinal_features + qualitative_features)), values_orders=values_orders
    )
    x_discretized = discretizer.fit_transform(x_train)

    expected = {'2': [2, '2'],
    '4': [4, '4'],
    '3': [3, '3'],
    '7': [7, '7'],
    '1': [1, '1'],
    '5': [5, '5'],
    '6': [6, '6']}
    assert discretizer.values_orders['Discrete_Qualitative_noorder'].contained == expected, "Not correctly converted for qualitative with integers"

    expected = {'2': [2.0, '2'],
    '4': [4.0, '4'],
    '3': [3.0, '3'],
    '1': [1.0, '1'],
    '5': [5.0, '5'],
    '6': [6.0, '6']}
    assert discretizer.values_orders['Discrete_Qualitative_lownan_noorder'].contained == expected, "Not correctly converted for qualitative with integers and nans"

    expected = {'2': [2.0, '2'],
    '4': [4.0, '4'],
    '3': [3.0, '3'],
    '0.5': [0.5, '0.5'],
    '1': [1.0, '1'],
    '5': [5.0, '5'],
    '6': [6.0, '6']}
    assert discretizer.values_orders['Discrete_Qualitative_rarevalue_noorder'].contained == expected, "Not correctly converted for qualitative with integers and floats"

    expected = {'Low-': ['Low-'],
    'Low': ['Low'],
    'Low+': ['Low+'],
    'Medium-': ['Medium-'],
    'Medium': ['Medium'],
    'Medium+': ['Medium+'],
    'High-': ['High-'],
    'High': ['High'],
    'High+': ['High+']}
    assert discretizer.values_orders['Qualitative_Ordinal_lownan'].contained == expected, "No conversion for already string features"
    
    expected = {'Low-': ['Low-'],
    'Low': ['Low'],
    'Low+': ['Low+'],
    'Medium-': ['Medium-'],
    'Medium': ['Medium'],
    'Medium+': ['Medium+'],
    'High-': ['High-'],
    'High': ['High'],
    'High+': ['High+']}
    assert discretizer.values_orders['Qualitative_Ordinal'].contained == expected, "No conversion for not specified featues"

    expected = {'1': ['1'], '2': [2.0, '2'], '3': [3.0, '3'], '4': [4.0, '4'], '5': [5.0, '5'], '6': [6.0, '6'], '7': [7.0, '7']}
    assert discretizer.values_orders['Discrete_Qualitative_highnan'].contained == expected, "Original order should be kept for ordinal features"



"""Set of tests for feature_selector module."""

from AutoCarver.feature_selector import *
from pytest import fixture


def test_feature_selector_quanti():
    assert False, "TODO"

def test_feature_selector_quali():
    assert False, "TODO"



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from typing import Any, Dict, List, Tuple

import numpy as np
from IPython.display import display_html  # TODO: remove this
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.pyplot import subplots
from matplotlib.ticker import PercentFormatter
from pandas import DataFrame, Series, crosstab, unique
from scipy.stats import chi2_contingency
from seaborn import color_palette, despine
from tqdm import tqdm

from .discretizers.discretizers import Discretizer
from .discretizers.utils.base_discretizers import (
    GroupedList,
    GroupedListDiscretizer,
    convert_to_labels,
    convert_to_values,
    is_equal,
)


# TODO: display tables
class AutoCarver(GroupedListDiscretizer):
    """Automatic carving of continuous, categorical and categorical ordinal
    features that maximizes association with a binary target.

    Modalities/values of features are carved/regrouped according to a computed
    specific order defined based on their types:
     - [Qualitative features] grouped based on modality target rate.
     - [Qualitative ordinal features] grouped based on specified modality order
     - [Quantitative features] grouped based on the order of their values.
    Uses Tschurpow's T to find the optimal carving (regrouping) of modalities/
    values of features.

    Parameters
    ----------
    values_orders: dict, default {}
        Dictionnary of features and list of their respective values' order.
        Exemple: for an `age` feature, `values_orders` could be
        `{'age': ['0-18', '18-30', '30-50', '50+']}`.

    sort_by: str, default 'tschuprowt'
        Association measure used to find the optimal group modality combination
        Implemented: ['cramerv', 'tschuprowt']

    max_n_mod: int, default 5
        Maximum number of modalities for the carved features (excluding `nan`).
         - All possible combinations of less than `max_n_mod` groups of
           modalities will be tested.
        Recommandation: `max_n_mod` should be set from 4 (faster) to 6
        (preciser).

    keep_nans: bool, default False
        Whether or not to group `numpy.nan` to other modalities/values.

    Examples
    ----------

    from AutoCarver import AutoCarver
    from Discretizers import Discretizer
    from sklearn.pipeline import Pipeline

    # defining training and testing sets
    X_train, y_train = train_set, train_set[target]
    X_test, y_test = test_set, test_set[target]

    # specifying features to be carved
    selected_quanti = ['amount', 'distance', 'length', 'height']
    selected_quali = ['age', 'type', 'grade', 'city']

    # specifying orders of categorical ordinal features
    values_orders = {
        'age': ['0-18', '18-30', '30-50', '50+'],
        'grade': ['A', 'B', 'C', 'D', 'J', 'K', 'NN']
    }

    # pre-processing of features into categorical ordinal features
    discretizer = Discretizer(
        selected_quanti, selected_quali, min_freq=0.02, q=20,
        values_orders=values_orders
    )
    X_train = discretizer.fit_transform(X_train, y_train)
    X_test = discretizer.transform(X_test)

    # storing Discretizer
    pipe = [('Discretizer', discretizer)]

    # updating features' values orders (every features are qualitative ordinal)
    values_orders = discretizer.values_orders

    # intiating AutoCarver
    auto_carver = AutoCarver(
        values_orders, sort_by='cramerv', max_n_mod=5, sample_size=0.01)

    # fitting on training sample
    # a test sample can be specified to evaluate carving robustness
    X_train = auto_carver.fit_transform(X_train, y_train, X_test, y_test)

    # applying transformation on test sample
    X_test = auto_carver.transform(X_test)

    # identifying non stable/robust features
    print(auto_carver.non_viable_features)

    # storing fitted GroupedListDiscretizer in a sklearn.pipeline.Pipeline
    pipe += [('AutoCarver', auto_carver)]
    pipe = Pipeline(pipe)

    # applying pipe to a validation set or in production
    X_val = pipe.transform(X_val)

    """

    def __init__(
        self,
        quantitative_features: List[str],
        qualitative_features: List[str],
        min_freq: float,
        *,
        ordinal_features: List[str] = None,
        values_orders: Dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        min_carved_freq: float = 0,  # TODO: update this parameter so that it is set according to frequency rather than number of groups
        sort_by: str = "tschuprowt",
        str_nan: str = "__NAN__",
        str_default: str = "__OTHER__",
        output_dtype: str = 'float',
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = True,
    ) -> None:
        """_summary_

        Parameters
        ----------
        quantitative_features : List[str]
            _description_
        qualitative_features : List[str]
            _description_
        min_freq : float
            _description_
        ordinal_features : List[str], optional
            _description_, by default None
        values_orders : Dict[str, GroupedList], optional
            _description_, by default None
        max_n_mod : int, optional
            _description_, by default 5
        min_carved_freq : float, optional
            _description_, by default 0
        str_nan : str, optional
            _description_, by default "__NAN__"
        str_default : str, optional
            _description_, by default "__OTHER__"
        output_dtype : str, optional
            _description_, by default 'float'
        dropna : bool, optional
            _description_, by default True
        copy : bool, optional
            _description_, by default False
        verbose : bool, optional
            _description_, by default True
        """
        # Lists of features
        self.features = list(set(quantitative_features + qualitative_features + ordinal_features))
        if ordinal_features is None:
            ordinal_features = []
        self.ordinal_features = list(set(ordinal_features))

        # initializing input_dtypes
        self.input_dtypes = {feature: "str" for feature in qualitative_features + ordinal_features}
        self.input_dtypes.update({feature: "float" for feature in quantitative_features})

        # Initiating GroupedListDiscretizer
        super().__init__(
            features=self.features,
            values_orders=values_orders,
            input_dtypes=self.input_dtypes,
            output_dtype=output_dtype,
            str_nan=str_nan,
            copy=copy,
        )

        # class specific attributes
        self.min_freq = min_freq  # minimum frequency per base bucket
        self.max_n_mod = max_n_mod  # maximum number of modality per feature
        self.dropna = dropna  # whether or not to group NaNs with other modalities
        self.verbose = verbose
        self.str_default = str_default
        self.min_carved_freq = min_carved_freq
        self.min_group_size = 1
        measures = ["tschuprowt", "cramerv"]  # association measure used to find the best groups
        assert (
            sort_by in measures
        ), f"""Measure '{sort_by}' not yet implemented. Choose from: {str(measures)}."""
        self.sort_by = sort_by

    def prepare_data(
        self,
        X: DataFrame,
        y: Series,
        X_test: DataFrame = None,
        y_test: Series = None,
    ) -> Tuple[DataFrame, DataFrame]:
        """Checks validity of provided data

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_
        X_test : DataFrame, optional
            _description_, by default None
        y_test : Series, optional
            _description_, by default None

        Returns
        -------
        Tuple[DataFrame, DataFrame]
            Copies of (X, X_test)
        """
        # Checking for binary target and copying X
        x_copy = super().prepare_data(X, y)
        x_test_copy = super().prepare_data(X_test, y_test)

        return x_copy, x_test_copy

    def remove_feature(self, feature: str) -> None:
        """Removes a feature from all instances 

        Parameters
        ----------
        feature : str
            Column name
        """        
        if feature in self.features:
            super().remove_feature(feature)
            if feature in self.ordinal_features:
                self.ordinal_features.remove(feature)

    def fit(
        self,
        X: DataFrame,
        y: Series,
        X_test: DataFrame = None,
        y_test: Series = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_
        X_test : DataFrame, optional
            _description_, by default None
        y_test : Series, optional
            _description_, by default None
        """
        # preparing datasets and checking for wrong values
        x_copy, x_test_copy = self.prepare_data(X, y, X_test, y_test)

        # discretizing all features
        discretizer = Discretizer(
            quantitative_features=self.quantitative_features,
            qualitative_features=self.qualitative_features,
            min_freq=self.min_freq,
            ordinal_features=self.ordinal_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            str_default=self.str_default,
            copy=False,
            verbose=self.verbose,
        )
        x_copy = discretizer.fit_transform(x_copy, y)
        if x_test_copy is not None:
            x_test_copy = discretizer.transform(x_test_copy, y_test)
        self.input_dtypes.update(discretizer.input_dtypes)  # saving data types

        # updating values_orders according to base bucketization
        self.values_orders.update(discretizer.values_orders)

        # removing dropped features
        for feature in self.features:
            if feature not in discretizer.values_orders:
                self.remove_feature(feature)

        # converting potential quantiles into there respective labels
        labels_orders = convert_to_labels(
            features=self.features,
            quantitative_features=self.quantitative_features,
            values_orders=self.values_orders,
            str_nan=self.str_nan,
            dropna=False,
        )

        # computing crosstabs for each feature on train/test
        xtabs = get_xtabs(self.features, x_copy, y, labels_orders)
        xtabs_test = get_xtabs(self.features, x_test_copy, y_test, labels_orders)

        # optimal butcketization/carving of each feature
        for n, feature in enumerate(self.features):
            if self.verbose:  # verbose if requested
                print(f"\n---\n[AutoCarver] Fit {feature} ({n+1}/{len(self.features)})")

            # getting xtabs on train/test
            xtab = xtabs[feature]
            xtab_test = xtabs_test[feature]
            if self.verbose:  # verbose if requested
                print(xtab)

            # ordering
            order = labels_orders[feature]

            # getting best combination
            best_combination = self.get_best_combination(order, xtab, xtab_test=xtab_test)

            # checking that a suitable combination has been found
            if best_combination is not None:
                order, xtab, xtab_test = best_combination
                if self.verbose:  # verbose if requested
                    print(xtab)

                # updating label_orders
                labels_orders.update({feature: order})
            
            # no suitable combination has been found -> removing feature
            else:
                print(f"No robust combination for feature '{feature}' could be found. It will be ignored. You might have to increase the size of your test sample (test sample not representative of test sample for this feature) or you should consider dropping this features.")
                self.remove_feature(feature)
                if feature in labels_orders:
                    labels_orders.pop(feature)


        # converting potential labels into there respective values (quantiles)
        self.values_orders.update(
            convert_to_values(
                features=self.features,
                quantitative_features=self.quantitative_features,
                values_orders=self.values_orders,
                label_orders=labels_orders,
                str_nan=self.str_nan,
            )
        )

        # TODO pretty displaying

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self

    def get_best_combination(
        self,
        order: GroupedList,
        xtab: DataFrame,
        *,
        xtab_test: DataFrame = None,
    ) -> Tuple[GroupedList, DataFrame, DataFrame]:
        # raw ordering
        raw_order = GroupedList(order)
        if self.str_nan in raw_order:
            raw_order.remove(self.str_nan)

        # filtering out nans if requested from train/test crosstabs
        raw_xtab = filter_nan_xtab(xtab, self.str_nan)
        raw_xtab_test = filter_nan_xtab(xtab_test, self.str_nan)

        # checking for non-nan values
        best_association = None
        if raw_xtab.shape[0] > 1:
            # all possible consecutive combinations
            combinations = consecutive_combinations(
                raw_order, self.max_n_mod, min_group_size=self.min_group_size
            )

            # getting most associated combination
            best_association = get_best_association(
                raw_xtab,
                combinations,
                sort_by=self.sort_by,
                xtab_test=raw_xtab_test,
                verbose=self.verbose,
            )

            # applying best_combination to order and xtabs
            if best_association is not None:
                order = order_apply_combination(order, best_association["combination"])
                xtab = xtab_apply_order(xtab, order)
                xtab_test = xtab_apply_order(xtab_test, order)

            # grouping NaNs if requested to drop them (dropna=True)
            if self.dropna and self.str_nan in order and best_association is not None:
                # raw ordering without nans
                raw_order = GroupedList(order)
                raw_order.remove(self.str_nan)

                # all possible consecutive combinations
                combinations = consecutive_combinations(
                    raw_order, self.max_n_mod, min_group_size=self.min_group_size
                )

                # adding combinations with NaNs
                nan_combinations = add_nan_in_combinations(
                    combinations, self.str_nan, self.max_n_mod
                )

                # getting most associated combination
                best_association = get_best_association(
                    xtab,
                    nan_combinations,
                    sort_by=self.sort_by,
                    xtab_test=xtab_test,
                    verbose=self.verbose,
                )

                # applying best_combination to order and xtab
                if best_association is not None:
                    order = order_apply_combination(order, best_association["combination"])
                    xtab = xtab_apply_order(xtab, order)
                    xtab_test = xtab_apply_order(xtab_test, order)

        # checking that a suitable combination has been found
        if best_association is not None:
            return order, xtab, xtab_test

    def display_xtabs(
        self,
        feature: str,
        caption: str,
        xtab: DataFrame,
        xtab_test: DataFrame = None,
    ) -> None:
        """Pretty display of frequency and target rate per modality on the same line."""

        # known_order per feature
        known_order = self.values_orders[feature]

        # target rate and frequency on TRAIN
        train_stats = stats_xtab(xtab, known_order)

        # target rate and frequency on TEST
        if xtab_test is not None:
            test_stats = stats_xtab(xtab_test, train_stats.index, train_stats.labels)
            test_stats = test_stats.set_index("labels")  # setting labels as indices

        # setting TRAIN labels as indices
        train_stats = train_stats.set_index("labels")

        # Displaying TRAIN modality level stats
        train_style = train_stats.style.background_gradient(cmap="coolwarm")  # color scaling
        train_style = train_style.set_table_attributes(
            "style='display:inline'"
        )  # printing in notebook
        train_style = train_style.set_caption(f"{caption} distribution on X:")  # title
        html = train_style._repr_html_()

        # adding TEST modality level stats
        if xtab_test is not None:
            test_style = test_stats.style.background_gradient(cmap="coolwarm")  # color scaling
            test_style = test_style.set_table_attributes(
                "style='display:inline'"
            )  # printing in notebook
            test_style = test_style.set_caption(f"{caption} distribution on X_test:")  # title
            html += " " + test_style._repr_html_()

        # displaying html of colored DataFrame
        display_html(html, raw=True)


def stats_xtab(
    xtab: DataFrame,
    known_order: List[Any] = None,
    known_labels: List[Any] = None,
) -> DataFrame:
    """Computes column (target) rate per row (modality) and row frequency"""

    # target rate and frequency statistics per modality
    stats = DataFrame(
        {
            # target rate per modality
            "target_rate": xtab[1].divide(xtab.sum(axis=1)),
            # frequency per modality
            "frequency": xtab.sum(axis=1) / xtab.sum().sum(),
        }
    )

    # sorting statistics
    # case 0: default ordering based on observed target rate
    if known_order is None:
        order = list(stats.sort_values("target_rate", ascending=True).index)

    # case 1: a known_order was provided
    else:
        order = known_order[:]

    # modalities' labels
    # case 0: default labels
    if known_labels is None:
        # accessing string representation of the GroupedList
        if isinstance(known_order, GroupedList):
            labels = known_order.get_repr()

        # labels are the default order
        else:
            labels = order[:]

    # case 1: known_labels were provided
    else:
        labels = known_labels[:]

    # keeping values missing from the order at the end
    unknown_modality = [mod for mod in xtab.index if mod not in order]
    for mod in unknown_modality:
        order = [c for c in order if not is_equal(c, mod)] + [mod]
        labels = [c for c in labels if not is_equal(c, mod)] + [mod]

    # sorting statistics
    stats = stats.reindex(order, fill_value=0)
    stats["labels"] = labels

    return stats


def filter_nan_xtab(xtab: DataFrame, str_nan: str) -> DataFrame:
    """Filters out nans from the crosstab"""

    # cehcking for values in crosstab
    filtered_xtab = None
    if xtab is not None:
        # filtering out nans if requested from train crosstab
        filtered_xtab = xtab.copy()
        if str_nan in xtab.index:
            filtered_xtab = xtab.drop(str_nan, axis=0)

    return filtered_xtab


def get_xtabs(
    features: List[str], X: DataFrame, y: Series, labels_orders: Dict[str, GroupedList]
) -> Dict[str, DataFrame]:
    """Computes crosstabs for specified features and ensures that the crosstab is ordered according to the known labels"""

    # checking for empty datasets
    xtabs = {feature: None for feature in features}
    if X is not None:
        # crosstab for each feature
        for feature in features:
            # computing crosstab with str_nan
            xtab = crosstab(X[feature], y)

            # reordering according to known_order
            xtab = xtab.reindex(labels_orders[feature])  # TODO: fill nans for x_test?

            # storing results
            xtabs.update({feature: xtab})

    return xtabs


def association_xtab(xtab: DataFrame, n_obs, n_mod_y) -> Dict[str, float]:
    """Computes measures of association between feature x and feature2."""

    # number of values taken by the features
    n_mod_x = xtab.shape[0]

    # Chi2 statistic
    chi2 = chi2_contingency(xtab)[0]

    # Cramer's V
    cramerv = np.sqrt(chi2 / n_obs / (n_mod_y - 1))

    # Tschuprow's T
    tschuprowt = np.sqrt(chi2 / n_obs / np.sqrt((n_mod_x - 1) * (n_mod_y - 1)))

    return {"cramerv": cramerv, "tschuprowt": tschuprowt}


def vectorized_groupby_sum(xtab: DataFrame, groupby: List[str]):
    """Groups a crosstab by groupby and sums column values by groups"""

    # all indices that may be duplicated
    index_values = np.array(groupby)

    # all unique indices deduplicated
    unique_indices = np.unique(index_values)

    # initiating summed up array with zeros
    summed_values = np.zeros((len(unique_indices), len(xtab.columns)))

    # for each unique_index found in index_values sums xtab.Values at corresponding position in summed_values
    np.add.at(summed_values, np.searchsorted(unique_indices, index_values), xtab.values)

    # converting back to dataframe
    grouped_xtab = DataFrame(summed_values, index=unique_indices, columns=xtab.columns)

    return grouped_xtab


def combinations_at_index(start_idx, order, nb_remaining_groups, min_group_size=1):
    """Gets all possible combinations of sizes up to the last element of a list"""

    # iterating over each possible length of groups
    for size in range(min_group_size, len(order) + 1):
        next_idx = start_idx + size  # index from which to start the next group

        # checking that next index is not off the order list
        if next_idx < len(order) + 1:
            # checking that there are remaining groups or that it is the last group
            if (nb_remaining_groups > 1) | (next_idx == len(order)):
                combination = list(order[start_idx:next_idx])
                yield (combination, next_idx, nb_remaining_groups - 1)


def consecutive_combinations(
    raw_order,
    max_group_size,
    min_group_size=1,
    nb_remaining_group=None,
    current_combination=None,
    next_index=None,
    all_combinations=None,
):
    """Computes all possible combinations of values of order up to max_group_size."""

    # initiating recursive attributes
    if current_combination is None:
        current_combination = []
    if next_index is None:
        next_index = 0
    if nb_remaining_group is None:
        nb_remaining_group = max_group_size
    if all_combinations is None:
        all_combinations = []

    # getting combinations for next index
    next_combinations = [
        elt
        for elt in combinations_at_index(next_index, raw_order, nb_remaining_group, min_group_size)
    ]

    # stop case: no next_combinations possible -> adding to all_combinations
    if len(next_combinations) == 0 and min_group_size < len(current_combination) <= max_group_size:
        # saving up combination
        all_combinations += [current_combination]

        # resetting remaining number of groups
        nb_remaining_group = max_group_size

    # otherwise: adding all next_combinations to the current_combination
    for combination, next_index, current_nb_remaining_group in next_combinations:
        # going a rank further in the raw_xtab
        consecutive_combinations(
            raw_order,
            max_group_size,
            min_group_size=min_group_size,
            nb_remaining_group=current_nb_remaining_group,
            current_combination=current_combination + [combination],
            next_index=next_index,
            all_combinations=all_combinations,
        )

    return all_combinations


def xtab_target_rate(xtab: DataFrame) -> DataFrame:
    """Computes target rate per row for a binary target (column) in a crosstab"""

    return xtab[1].divide(xtab[0]).sort_values()


def get_best_association(
    xtab: DataFrame,
    combinations: List[List[str]],
    sort_by: str,
    xtab_test: DataFrame = None,
    verbose: bool = False,
) -> Dict[str, Any]:
    """Computes associations of the xtab for each combination"""

    # values to groupby indices with
    indices_to_groupby = [
        [value for values in ([group[0]] * len(group) for group in combination) for value in values]
        for combination in combinations
    ]

    # grouping xtab by its indices
    grouped_xtabs = [
        vectorized_groupby_sum(xtab, index_to_groupby)
        for index_to_groupby in tqdm(
            indices_to_groupby, disable=not verbose, desc="Grouping modalities   "
        )
    ]

    # computing associations for each xtabs
    n_obs = xtab.sum().sum()  # number of observation
    n_mod_y = len(xtab.columns)  # number of modalities and minimum number of modalities
    associations_xtab = [
        association_xtab(grouped_xtab, n_obs, n_mod_y)
        for grouped_xtab in tqdm(grouped_xtabs, disable=not verbose, desc="Computing associations")
    ]

    # adding corresponding combination to the association
    for combination, index_to_groupby, association, grouped_xtab in zip(
        combinations, indices_to_groupby, associations_xtab, grouped_xtabs
    ):
        association.update(
            {"combination": combination, "index_to_groupby": index_to_groupby, "xtab": grouped_xtab}
        )

    # sorting associations according to specified metric
    associations_xtab = (
        DataFrame(associations_xtab).sort_values(sort_by, ascending=False).to_dict(orient="records")
    )

    # case 0: no test sample provided -> not testing for robustness
    if xtab_test is None:
        return associations_xtab[0]

    # case 1: testing viability on provided test sample
    for association in tqdm(associations_xtab, disable=not verbose, desc="Testing robustness    "):
        # needed parameters
        index_to_groupby, xtab = (
            association["index_to_groupby"],
            association["xtab"],
        )

        # grouping rows of the test crosstab
        grouped_xtab_test = vectorized_groupby_sum(xtab_test, index_to_groupby)

        # computing target rate ranks per value
        train_ranks = xtab_target_rate(xtab).index
        test_ranks = xtab_target_rate(grouped_xtab_test).index

        # viable on test sample: grouped values have the same ranks in train/test
        if all(train_ranks == test_ranks):
            return association


def add_nan_in_combinations(
    combinations: List[List[str]], str_nan: str, max_n_mod: int
) -> List[List[str]]:
    """Adds nan to each possible group and a last group only with nan if the max_n_mod is not reached by the combination"""

    # iterating over each combination
    nan_combinations = []
    for combination in combinations:
        # adding nan to each group of the combination
        nan_combination = []
        for n in range(len(combination)):
            # copying input combination
            new_combination = combination[:]
            # adding nan to the nth group
            new_combination[n] = new_combination[n] + [str_nan]
            # storing updated combination with attributed group to nan
            nan_combination += [new_combination]

        # if max_n_mod is not reached adding a combination with nans alone
        if len(combination) < max_n_mod:
            # copying input combination
            new_combination = combination[:]
            # adding a group for nans only
            nan_combination += [new_combination + [[str_nan]]]

        nan_combinations += nan_combination

    return nan_combinations


def order_apply_combination(order: GroupedList, combination: List[List[Any]]) -> GroupedList:
    """Converts a list of combination to a GroupedList"""

    order_copy = GroupedList(order)
    for combi in combination:
        order_copy.group_list(combi, combi[0])

    return order_copy


def xtab_apply_order(xtab: DataFrame, order: GroupedList) -> DataFrame:
    """Applies an order (combination) to a crosstab

    Parameters
    ----------
    xtab : DataFrame
        Crosstab
    order : GroupedList
        Combination of index to apply to the crosstab

    Returns
    -------
    Dict[str, Any]
        _description_
    """
    # checking for input values
    combi_xtab = None
    if xtab is not None:
        # grouping modalities in the crosstab
        groups = list(map(order.get_group, xtab.index))
        combi_xtab = xtab.groupby(groups, dropna=False, sort=False).sum()

    return combi_xtab


def plot_stats(stats: DataFrame) -> Tuple[Figure, Axes]:
    """Barplot of the volume and target rate"""

    x = [0] + [elt for e in stats["frequency"].cumsum()[:-1] for elt in [e] * 2] + [1]
    y2 = [elt for e in list(stats["target_rate"]) for elt in [e] * 2]
    s = list(stats.index)
    scaled_y2 = [(y - min(y2)) / (max(y2) - min(y2)) for y in y2]
    c = color_palette("coolwarm", as_cmap=True)(scaled_y2)

    fig, ax = subplots()

    for i in range(len(stats)):
        k = i * 2
        ax.fill_between(x[k : k + 2], [0, 0], y2[k : k + 2], color=c[k])
        ax.text(
            sum(x[k : k + 2]) / 2,
            y2[k],
            s[i],
            rotation=90,
            ha="center",
            va="bottom",
        )

    ax.xaxis.set_major_formatter(PercentFormatter(xmax=1))
    ax.yaxis.set_major_formatter(PercentFormatter(xmax=1))
    ax.set_xlabel("Volume")
    ax.set_ylabel("Target rate")
    ax.set_xlim(0, 1)
    ax.set_ylim(0)
    despine()

    return fig, ax



"""Set of tests for discretizers module."""

from AutoCarver.discretizers import discretizers
from pandas import DataFrame
from numpy import inf

def test_quantitative_discretizer(x_train: DataFrame):
    """Tests QuantitativeDiscretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    """

    features = ['Quantitative', 'Discrete_Quantitative', 'Discrete_Quantitative_highnan', 'Discrete_Quantitative_lownan', 'Discrete_Quantitative_rarevalue']
    min_freq = 0.1

    discretizer = discretizers.QuantitativeDiscretizer(features, min_freq=min_freq)
    x_discretized = discretizer.fit_transform(x_train, x_train["quali_ordinal_target"])

    assert '__NAN__' in discretizer.values_orders["Discrete_Quantitative_lownan"], "Missing order should not be grouped with ordinal_discretizer"
    assert all(x_discretized.Quantitative.value_counts(normalize=True) >= min_freq), "Non-nan value was not grouped"
    assert discretizer.values_orders["Discrete_Quantitative_lownan"] == [1.0, 2.0, 3.0, inf, '__NAN__'], "NaNs should not be grouped whatsoever"
    assert discretizer.values_orders["Discrete_Quantitative_rarevalue"] == [1.0, 2.0, 3.0, inf], "Rare values should be grouped to the closest one (OrdinalDiscretizer)"


def test_qualitative_discretizer(x_train: DataFrame):
    """Tests QualitativeDiscretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    """


    features = ["Qualitative", "Qualitative_grouped", "Qualitative_lownan", "Qualitative_highnan"]
    ordinal_features = ["Qualitative_Ordinal", "Qualitative_Ordinal_lownan"]
    values_orders = {
        "Qualitative_Ordinal": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Qualitative_Ordinal_lownan": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
    }

    min_freq = 0.1

    discretizer = discretizers.QualitativeDiscretizer(features, min_freq, ordinal_features=ordinal_features, values_orders=values_orders, copy=True, verbose=True)
    x_discretized = discretizer.fit_transform(x_train, x_train["quali_ordinal_target"])


    quali_expected = {
        '__OTHER__': ['Category A', 'Category D', 'Category F', '__OTHER__'],
         'Category C': ['Category C'],
         'Category E': ['Category E'],
    }
    assert discretizer.values_orders['Qualitative'].contained == quali_expected, "Values less frequent than min_freq should be grouped into default_value"
    quali_lownan_expected = {
        '__NAN__': ['__NAN__'],
        '__OTHER__': ['Category D', 'Category F', '__OTHER__'],
        'Category C': ['Category C'],
        'Category E': ['Category E'],
    }
    assert discretizer.values_orders['Qualitative_lownan'].contained == quali_lownan_expected, "If any, NaN values should be put into str_nan and kept by themselves"

    expected_ordinal = {
        'Low+': ['Low-', 'Low', 'Low+'],
        'Medium-': ['Medium-'],
        'Medium': ['Medium'],
        'High': ['Medium+', 'High-', 'High'],
        'High+': ['High+']
    }
    expected_ordinal_lownan = {
        'Low+': ['Low', 'Low-', 'Low+'],
        'Medium-': ['Medium-'],
        'Medium': ['Medium'],
        'High': ['Medium+', 'High-', 'High'],
        'High+': ['High+'],
        '__NAN__': ['__NAN__']
    }
    assert discretizer.values_orders['Qualitative_Ordinal'].contained == expected_ordinal, "Values not correctly grouped"
    assert discretizer.values_orders['Qualitative_Ordinal_lownan'].contained == expected_ordinal_lownan, "NaNs should stay by themselves."

def test_discretizer(x_train: DataFrame, x_test_1: DataFrame):
    """Tests Discretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    x_test_1 : DataFrame
        Simulated Test DataFrame
    """

    quantitative_features = ['Quantitative', 'Discrete_Quantitative_highnan', 'Discrete_Quantitative_lownan', 'Discrete_Quantitative', 'Discrete_Quantitative_rarevalue']
    qualitative_features = ["Qualitative", "Qualitative_grouped", "Qualitative_lownan", "Qualitative_highnan", "Discrete_Qualitative_noorder", "Discrete_Qualitative_lownan_noorder", "Discrete_Qualitative_rarevalue_noorder"]
    ordinal_features = ["Qualitative_Ordinal", "Qualitative_Ordinal_lownan", "Discrete_Qualitative_highnan"]
    values_orders = {
        "Qualitative_Ordinal": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Qualitative_Ordinal_lownan": ['Low-', 'Low', 'Low+', 'Medium-', 'Medium', 'Medium+', 'High-', 'High', 'High+'],
        "Discrete_Qualitative_highnan" : ["1", "2", "3", "4", "5", "6", "7"],
    }



    # minimum frequency per modality + apply(find_common_modalities) outputs a Series
    min_freq = 0.1

    # discretizing features
    discretizer = discretizers.Discretizer(
        quantitative_features=quantitative_features,
        qualitative_features=qualitative_features,
        ordinal_features=ordinal_features,
        values_orders=values_orders,
        min_freq=min_freq,
        copy=True,
    )
    x_discretized = discretizer.fit_transform(x_train, x_train["quali_ordinal_target"])
    x_test_discretized = discretizer.transform(x_test_1)



    assert all(x_discretized.Quantitative.value_counts(normalize=True) >= min_freq), "Non-nan value was not grouped"
    assert discretizer.values_orders["Discrete_Quantitative_lownan"] == [1.0, 2.0, 3.0, inf, '__NAN__'], "NaNs should not be grouped whatsoever"
    assert discretizer.values_orders["Discrete_Quantitative_rarevalue"] == [1.0, 2.0, 3.0, inf], "Rare values should be grouped to the closest one (OrdinalDiscretizer)"


    quali_expected = {
        '__OTHER__': ['Category A', 'Category D', 'Category F', '__OTHER__'],
        'Category C': ['Category C'],
        'Category E': ['Category E'],
    }
    assert discretizer.values_orders['Qualitative'].contained == quali_expected, "Values less frequent than min_freq should be grouped into default_value"
    quali_lownan_expected = {
        '__NAN__' : ['__NAN__'],
        '__OTHER__': ['Category D', 'Category F', '__OTHER__'],
        'Category C': ['Category C'],
        'Category E': ['Category E'],
    }
    assert discretizer.values_orders['Qualitative_lownan'].contained == quali_lownan_expected, "If any, NaN values should be put into str_nan and kept by themselves"

    expected_ordinal = {
        'Low+': ['Low-', 'Low', 'Low+'],
        'Medium-': ['Medium-'],
        'Medium': ['Medium'],
        'High': ['Medium+', 'High-', 'High'],
        'High+': ['High+']
    }
    expected_ordinal_lownan = {
        'Low+': ['Low', 'Low-', 'Low+'],
        'Medium-': ['Medium-'],
        'Medium': ['Medium'],
        'High': ['Medium+', 'High-', 'High'],
        'High+': ['High+'],
        '__NAN__': ['__NAN__'],
    }
    assert discretizer.values_orders['Qualitative_Ordinal'].contained == expected_ordinal, "Values not correctly grouped"
    assert discretizer.values_orders['Qualitative_Ordinal_lownan'].contained == expected_ordinal_lownan, "NaNs should stay by themselves."

    # Testing out qualitative with int/float values inside -> StringDiscretizer
    expected = {
        '2': [2.0, '2'],
        '4': [4.0, '4'],
        '1': [1.0, '1'],
        '3': [3.0, '3'],
        '__OTHER__': [0.5, '0.5', 6.0, '6', 5.0, '5', '__OTHER__']
    }
    assert discretizer.values_orders["Discrete_Qualitative_rarevalue_noorder"].contained == expected, "Qualitative features with float values should be converted to string and there values stored in the values_orders"
    expected = {
        '2': [2, '2'],
        '4': [4, '4'],
        '1': [1, '1'],
        '3': [3, '3'],
        '__OTHER__': [7, '7', 6, '6', 5, '5', '__OTHER__']
    }
    assert discretizer.values_orders["Discrete_Qualitative_noorder"].contained == expected, "Qualitative features with int values should be converted to string and there values stored in the values_orders"
    expected = {
        '2': ['1', 2.0, '2'],
        '3': [3.0, '3'],
        '4': [4.0, '4'],
        '5': [6.0, '6', 7.0, '7', 5.0, '5'],
        '__NAN__': ['__NAN__']
    }
    assert discretizer.values_orders["Discrete_Qualitative_highnan"].contained == expected, "Ordinal qualitative features with int or float values that contain nan should be converted to string and there values stored in the values_orders"

    # checking for inconsistancies in tranform
    for feature in discretizer.features:
        test_unique = x_test_discretized[feature].unique()
        train_unique = x_discretized[feature].unique()
        assert all(value in test_unique for value in train_unique), "Missing value from test (at transform step)"
        assert all(value in train_unique for value in test_unique), "Missing value from train (at transform step)"



"""Tools to select the best Quantitative and Qualitative features."""

from math import sqrt
from random import shuffle
from typing import Any, Callable, Dict, Tuple

from IPython.display import display_html
from numpy import inf, nan, ones, triu
from pandas import DataFrame, Series, crosstab, notna
from scipy.stats import chi2_contingency, kruskal
from sklearn.base import BaseEstimator, TransformerMixin
from statsmodels.formula.api import ols
from statsmodels.stats.outliers_influence import variance_inflation_factor

from .discretizers.utils.base_discretizers import GroupedList, GroupedListDiscretizer

# TODO: convert to groupedlistdiscretizer
# TODO: add parameter to shut down displayed info
class FeatureSelector(BaseEstimator, TransformerMixin):
    """A pipeline of measures to perform EDA and feature pre-selection

     - best features are the n_best of each measure
     - selected features are stored in FeatureSelector.best_features

    Parameters
    ----------
    features: list[str]
        Features on which to compute association.
    n_best, int:
        Number of features to be selected
    sample_size: float, default 1.
        Should be set between ]0, 1]
        Size of sampled list of features speeds up computation.
        By default, all features are used. For sample_size=0.5,
        FeatureSelector will search for the best features in
        features[:len(features)//2] and then in features[len(features)//2:]
    measures, list[Callable]: default list().
        List of association measures to be used.
        Implemented measures are:
            [Quantitative Features]
             - For association evaluation: `kruskal_measure`, `R_measure`
             - For outlier detection: `zscore_measure`, `iqr_measure`
            [Qualitative Features]
             - For correlation: `chi2_measure`, `cramerv_measure`, `tschuprowt_measure`
        Ranks features based on last measure of the list.
    filters, list[Callable]: default list().
        List of filters to be used.
        Implemented filters are:
            [Quantitative Features]
             - For linear correlation: `spearman_filter`, `pearson_filter`
             - For multicoloinearity: `vif_filter`
            [Qualitative Features]
             - For correlation: `cramerv_filter`, `tschuprowt_filter`

    Thresholds (to be passed as kwargs)
    ----------
    thresh_measure, float: default 0.
        Minimum association between target and features
        To be used with: `measure_filter`
    name_measure, str
        Measure to be used for minimum association filtering
        To be used with: `measure_filter`
    thresh_nan, float: default 1.
        Maximum percentage of NaNs in a feature
        To be used with: `nans_measure`
    thresh_mode, float: default 1.
        Maximum percentage of the mode of a feature
        To be used with: `mode_measure`
    thresh_outlier, float: default 1.
        Maximum percentage of Outliers in a feature
        To be used with: `iqr_measure`, `zscore_measure`
    thresh_corr, float: default 1.
        Maximum association between features
        To be used with: `spearman_filter`, `pearson_filter`, `cramerv_filter`, `tschuprowt_filter`
    thresh_vif, float: default inf
        Maximum VIF between features
        To be used with: `vif_filter`
    ascending, bool default False
        According to this measure:
         - True: Lower values of the measure are to be considered as more associated to the target
         - False: Higher values of the measure are to be considered as more associated to the target
    """

    def __init__(
        self,
        features: list[str],
        n_best: int,
        measures: list[Callable],
        *,
        filters: list[Callable] = None,
        sample_size: float = 1.0,
        values_orders: dict[str, GroupedList] = None,
        copy: bool = True,
        drop: bool = False,  # TODO
        verbose: bool = True,
        **params,
    ) -> None:
        """_summary_

        Parameters
        ----------
        features : list[str]
            _description_
        n_best : int
            _description_
        measures : list[Callable], optional
            _description_, by default list()
        filters : list[Callable], optional
            _description_, by default list()
        sample_size : float, optional
            _description_, by default 1.0
        copy : bool, optional
            _description_, by default True
        verbose : bool, optional
            _description_, by default True
        """
        print("Warning: not fully optimized for package versions greater than 4.")

        self.features = list(set(features))
        self.n_best = n_best
        assert n_best <= len(features), "Must set n_best <= len(features)"
        self.best_features = features[:]
        self.sample_size = sample_size

        self.measures = [dtype_measure, nans_measure, mode_measure] + measures[:]
        if filters is None:
            filters = []
        self.filters = [thresh_filter] + filters[:]
        self.sort_measures = [measure.__name__ for measure in measures[::-1]]

        # Values_orders from GroupedListDiscretizer
        if values_orders is None:
            values_orders = {}
        self.values_orders = {feature: GroupedList(value) for feature, value in values_orders.items()}

        self.drop = drop
        self.copy = copy
        self.verbose = verbose
        self.params = params

        self.associations = None
        self.filtered_associations = None

    def measure(self, x: Series, y: Series) -> Dict[str, Any]:
        """Measures association between x and y"""

        passed = True  # measures keep going only if previous basic tests are passed
        association = {}

        # iterating over each measure
        for measure in self.measures:
            passed, association = measure(passed, association, x, y, **self.params)

        return association

    def measure_apply(self, X: DataFrame, y: Series, features: list[str]) -> None:
        """Measures association between columns of X and y

        Parameters
        ----------
        ascending, bool default False
            According to this measure:
             - True: Lower values of the measure are to be considered as more associated to the target
             - False: Higher values of the measure are to be considered as more associated to the target
        """

        # applying association measure to each column
        self.associations = X[features].apply(self.measure, y=y, result_type="expand", axis=0).T

        # filtering non association measure (pct_zscore, pct_iqr...)
        asso_measures = [c for c in self.associations if "_measure" in c]
        self.sort_measures = [c for c in self.sort_measures if c in asso_measures]

        # sorting statistics if an association measure was provided
        self.associations = self.associations.sort_values(
            self.sort_measures, ascending=self.params.get("ascending", False)
        )

    def filter_apply(self, X: DataFrame, sort_measure: str) -> DataFrame:
        """Filters out too correlated features (least relevant first)

        Parameters
        ----------
        ascending, bool default False
            According to this measure:
             - True: Lower values of the measure are to be considered as more associated to the target
             - False: Higher values of the measure are to be considered as more associated to the target
        """

        # ordering features by sort_by
        self.filtered_associations = self.associations.sort_values(
            sort_measure, ascending=self.params.get("ascending", False)
        )

        # applying successive filters
        for filtering in self.filters:
            # ordered filtering
            self.filtered_associations = filtering(X, self.filtered_associations, **self.params)

    def display_stats(self, association: DataFrame, caption: str) -> None:
        """EDA of fitted associations"""

        # appllying style
        subset = [c for c in association if "pct_" in c or "_measure" in c or "_filter" in c]
        style = association.style.background_gradient(cmap="coolwarm", subset=subset)
        style = style.set_table_attributes("style='display:inline'")
        style = style.set_caption(caption)
        display_html(style._repr_html_(), raw=True)

    def fit_features(self, X: DataFrame, y: Series, features: list[str], n_best: int) -> list[str]:
        """Selects the n_best features amongst the specified ones"""

        # initial computation of all association measures
        self.measure_apply(X, y, features)

        # displaying association measure
        if self.verbose:
            self.display_stats(self.associations, "Raw association")

        # iterating over each sort_measures
        # useful when measures hints to specific associations
        ranks = []
        for n, sort_measure in enumerate(self.sort_measures):
            # filtering by sort_measure
            self.filter_apply(X, sort_measure)
            ranks += [list(self.filtered_associations.index)]

            # displaying filtered out association measure
            if n == 0 and self.verbose and len(self.filters) > 1:
                self.display_stats(self.filtered_associations, "Filtered association")

        # retrieving the n_best features per each ranking
        best_features = []
        if len(self.sort_measures) > 0:
            best_features = [feature for rank in ranks for feature in rank[:n_best]]
            best_features = list(set(best_features))  # deduplicating

        return best_features

    def fit(self, X: DataFrame, y: Series) -> None:
        """Selects the n_best features"""

        # splitting features in chunks
        if self.sample_size < 1:
            # shuffling features to get random samples of features
            shuffle(self.features)

            # number of features per sample
            chunks = int(len(self.features) // (1 / self.sample_size))

            # splitting feature list in samples
            feature_samples = [
                self.features[chunks * i : chunks * (i + 1)]
                for i in range(int(1 / self.sample_size) - 1)
            ]

            # adding last sample with all remaining features
            feature_samples += [self.features[chunks * (int(1 / self.sample_size) - 1) :]]

            # iterating over each feature samples
            best_features = []
            for features in feature_samples:
                # fitting association on features
                best_features += self.fit_features(X, y, features, int(self.n_best // 2))

        # splitting in chunks not requested
        else:
            best_features = self.features[:]

        # final selection with all best_features selected
        self.best_features = self.fit_features(X, y, best_features, self.n_best)

        # ordering best_features according to their rank
        self.best_features = [
            f for f in self.filtered_associations.index if f in self.best_features
        ]

        # removing feature from values_orders
        dropped_features = [f for f in self.associations.index if f not in self.best_features]
        for feature in dropped_features:
            if feature in self.values_orders:
                self.values_orders.pop(feature)

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Drops the non-selected columns from `features`.

        Parameters
        ----------
        X : DataFrame
            Contains columns named in `features`
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            `X` without non-selected columns from `features`.
        """

        return X


# MEASURES
def nans_measure(
    active: bool,
    association: Dict[str, Any],
    x: Series,
    y: Series = None,
    **params,
) -> Tuple[bool, Dict[str, Any]]:
    """Measure of the percentage of NaNs

    Parameters
    ----------
    thresh_nan, float: default 1.
      Maximum percentage of NaNs in a feature
    """

    # whether or not tests where passed
    if active:
        nans = x.isnull()  # ckecking for nans
        pct_nan = nans.mean()  # Computing percentage of nans

        # updating association
        association.update({"pct_nan": pct_nan})

        # Excluding feature that have to many NaNs
        active = pct_nan < params.get("thresh_nan", 1.0)

    return active, association


def dtype_measure(
    active: bool,
    association: Dict[str, Any],
    x: Series,
    y: Series = None,
    **params,
) -> Tuple[bool, Dict[str, Any]]:
    """Gets dtype"""

    # updating association
    association.update({"dtype": x.dtype})

    return active, association


def mode_measure(
    active: bool,
    association: Dict[str, Any],
    x: Series,
    y: Series = None,
    **params,
) -> Tuple[bool, Dict[str, Any]]:
    """Measure of the percentage of the Mode

    Parameters
    ----------
    thresh_mode, float: default 1.
      Maximum percentage of the mode of a feature
    """

    # whether or not tests where passed
    if active:
        mode = x.mode(dropna=True).values[0]  # computing mode
        pct_mode = (x == mode).mean()  # Computing percentage of the mode

        # updating association
        association.update({"pct_mode": pct_mode, "mode": mode})

        # Excluding feature with too frequent modes
        active = pct_mode < params.get("thresh_mode", 1.0)

    return active, association


def kruskal_measure(
    active: bool, association: Dict[str, Any], x: Series, y: Series, **params
) -> Tuple[bool, Dict[str, Any]]:
    """Kruskal-Wallis statistic between x (quantitative) and y (binary)"""

    # check that previous steps where passed
    if active:
        nans = x.isnull()  # ckecking for nans

        # computation of Kruskal-Wallis statistic
        kw = kruskal(x[(~nans) & (y == 0)], x[(~nans) & (y == 1)])

        # updating association
        if kw:
            association.update({"kruskal_measure": kw[0]})

    return active, association


def R_measure(
    active: bool, association: Dict[str, Any], x: Series, y: Series, **params
) -> Tuple[bool, Dict[str, Any]]:
    """R of the linear regression of x (quantitative) by y (binary)"""

    # check that previous steps where passed
    if active:
        nans = x.isnull()  # ckecking for nans

        # grouping feature and target
        ols_df = DataFrame({"feature": x[~nans], "target": y[~nans]})

        # fitting regression of feature by target
        regression = ols("feature~C(target)", ols_df).fit()

        # updating association
        if regression.rsquared:
            if regression.rsquared >= 0:
                association.update({"R_measure": sqrt(regression.rsquared)})
            else:
                association.update({"R_measure": nan})

    return active, association


def zscore_measure(
    active: bool,
    association: Dict[str, Any],
    x: Series,
    y: Series = None,
    **params,
) -> Tuple[bool, Dict[str, Any]]:
    """Computes outliers based on the z-score

    Parameters
    ----------
    thresh_outlier, float: default 1.
      Maximum percentage of Outliers in a feature
    """

    # check that previous steps where passed for computational optimization
    if active:
        mean = x.mean()  # mean of the feature
        std = x.std()  # standard deviation of the feature
        zscore = (x - mean) / std  # zscore per observation

        # checking for outliers
        outliers = abs(zscore) > 3
        pct_zscore = outliers.mean()

        # updating association
        association.update(
            {
                "pct_zscore": pct_zscore,
                "min": x.min(),
                "max": x.max(),
                "mean": mean,
                "std": std,
            }
        )

        # Excluding feature with too frequent modes
        active = pct_zscore < params.get("thresh_outlier", 1.0)

    return active, association


def iqr_measure(
    active: bool,
    association: Dict[str, Any],
    x: Series,
    y: Series = None,
    **params,
) -> Tuple[bool, Dict[str, Any]]:
    """Computes outliers based on the inter-quartile range

    Parameters
    ----------
    thresh_outlier, float: default 1.
      Maximum percentage of Outliers in a feature
    """

    # check that previous steps where passed for computational optimization
    if active:
        q3 = x.quantile(0.75)  # 3rd quartile
        q1 = x.quantile(0.25)  # 1st quartile
        iqr = q3 - q1  # inter quartile range
        iqr_bounds = q1 - 1.5 * iqr, q3 + 1.5 * iqr  # bounds of the iqr range

        # checking for outliers
        outliers = ~x.between(*iqr_bounds)
        pct_iqr = outliers.mean()

        # updating association
        association.update({"pct_iqr": pct_iqr, "q1": q1, "median": x.median(), "q3": q3})

        # Excluding feature with too frequent modes
        active = pct_iqr < params.get("thresh_outlier", 1.0)

    return active, association


def chi2_measure(
    active: bool, association: Dict[str, Any], x: Series, y: Series, **params
) -> Tuple[bool, Dict[str, Any]]:
    """Chi2 Measure between two Series of qualitative features"""

    # check that previous steps where passed
    if active:
        # computing crosstab between x and y
        xtab = crosstab(x, y)

        # Chi2 statistic
        chi2 = chi2_contingency(xtab)[0]

        # updating association
        association.update({"chi2_measure": chi2})

    return active, association


def cramerv_measure(
    active: bool, association: Dict[str, Any], x: Series, y: Series, **params
) -> Tuple[bool, Dict[str, Any]]:
    """Carmer's V between two Series of qualitative features"""

    # check that previous steps where passed
    if active:
        # computing chi2
        if "chi2_measure" not in association:
            active, association = chi2_measure(active, association, x, y, **params)

        # numnber of observations
        n_obs = (notna(x) & notna(y)).sum()

        # number of values taken by the features
        n_mod_x, n_mod_y = x.nunique(), y.nunique()
        min_n_mod = min(n_mod_x, n_mod_y)

        # Chi2 statistic
        chi2 = association.get("chi2_measure")

        # Cramer's V
        cramerv = sqrt(chi2 / n_obs / (min_n_mod - 1))

        # updating association
        association.update({"cramerv_measure": cramerv})

    return active, association


def tschuprowt_measure(
    active: bool, association: Dict[str, Any], x: Series, y: Series, **params
) -> Tuple[bool, Dict[str, Any]]:
    """Tschuprow's T between two Series of qualitative features"""

    # check that previous steps where passed
    if active:
        # computing chi2
        if "chi2_measure" not in association:
            active, association = chi2_measure(active, association, x, y, **params)

        # numnber of observations
        n_obs = (notna(x) & notna(y)).sum()

        # number of values taken by the features
        n_mod_x, n_mod_y = x.nunique(), y.nunique()

        # Chi2 statistic
        chi2 = association.get("chi2_measure")

        # Tschuprow's T
        dof_mods = sqrt((n_mod_x - 1) * (n_mod_y - 1))
        tschuprowt = 0
        if dof_mods > 0:
            tschuprowt = sqrt(chi2 / n_obs / dof_mods)

        # updating association
        association.update({"tschuprowt_measure": tschuprowt})

    return active, association


# FILTERS
def thresh_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Filters out missing association measure (did not pass a threshold)"""

    # drops rows with nans
    associations = ranks.dropna(axis=0)

    return associations


def measure_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Filters out specified measure's lower ranks than threshold

    Parameters
    ----------
    thresh_measure, float: default 0.
        Minimum association between target and features
        To be used with: `association_filter`
    name_measure, str
        Measure to be used for minimum association filtering
        To be used with: `association_filter`
    """

    associations = ranks.copy()

    # drops rows with nans
    if "name_measure" in params:
        associations = ranks[ranks[params.get("name_measure")] > params.get("thresh_measure", 0.0)]

    return associations


def quantitative_filter(
    X: DataFrame, ranks: DataFrame, corr_measure: str, **params
) -> Dict[str, Any]:
    """Computes max association between X and X (quantitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
        Maximum association between features
    """

    # accessing the prefered order
    prefered_order = ranks.index

    # computing correlation between features
    X_corr = X[prefered_order].corr(corr_measure).abs()
    X_corr = X_corr.where(triu(ones(X_corr.shape), k=1).astype(bool))

    # initiating list of maximum association per feature
    associations = []

    # iterating over each feature by target association order
    for feature in prefered_order:
        # correlation with features more associated to the target
        corr_with_better_features = X_corr.loc[:feature, feature]

        # maximum correlation with a better feature
        corr_with, worst_corr = corr_with_better_features.agg(["idxmax", "max"])

        # dropping the feature if it was too correlated to a better feature
        if worst_corr > params.get("thresh_corr", 1):
            X_corr = X_corr.drop(feature, axis=0).drop(feature, axis=1)

        # kept feature: updating associations with this feature
        else:
            associations += [
                {
                    "feature": feature,
                    f"{corr_measure}_filter": worst_corr,
                    f"{corr_measure}_with": corr_with,
                }
            ]

    # formatting ouput to DataFrame
    associations = DataFrame(associations).set_index("feature")

    # applying filter on association
    associations = ranks.join(associations, how="right")

    return associations


def spearman_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Computes max Spearman between X and X (quantitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
      Maximum association between features
    """

    # applying quantitative filter with spearman correlation
    return quantitative_filter(X, ranks, "spearman", **params)


def pearson_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Computes max Pearson between X and X (quantitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
      Maximum association between features
    """

    # applying quantitative filter with spearman correlation
    return quantitative_filter(X, ranks, "pearson", **params)


def vif_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Computes Variance Inflation Factor (multicolinearity)

    Parameters
    ----------
    thresh_vif, float: default inf
      Maximum VIF between features
    """

    # accessing the prefered order
    prefered_order = ranks.index

    # initiating list association per feature
    associations = []

    # list of dropped features
    dropped = []

    # iterating over each column
    for i, feature in enumerate(prefered_order):
        # identifying remaining more associated features
        better_features = [f for f in prefered_order[: i + 1] if f not in dropped]

        X_vif = X[better_features]  # keeping only better features
        X_vif = X_vif.dropna(axis=0)  # dropping NaNs for OLS

        # computation of VIF
        vif = nan
        if len(better_features) > 1 and len(X_vif) > 0:
            vif = variance_inflation_factor(X_vif.values, len(better_features) - 1)

        # dropping the feature if it was too correlated to a better feature
        if vif > params.get("thresh_vif", inf) and notna(vif):
            dropped += [feature]

        # kept feature: updating associations with this feature
        else:
            associations += [{"feature": feature, "vif_filter": vif}]

    # formatting ouput to DataFrame
    associations = DataFrame(associations).set_index("feature")

    # applying filter on association
    associations = ranks.join(associations, how="right")

    return associations


def qualitative_worst_corr(
    X: DataFrame,
    feature: str,
    ranks: DataFrame,
    corr_measure: Callable,
    **params,
):
    """Computes maximum association between a feature and features
    more associated to the target (according to ranks)
    """

    # measure name
    measure_name = corr_measure.__name__
    measure = measure_name.replace("_measure", "")

    # initiating worst correlation
    worst_corr = {"feature": feature}

    # features more associated with target
    better_features = list(ranks.loc[:feature].index)[:-1]

    # iterating over each better feature
    for better_feature in better_features:
        # computing association with better feature
        _, association = corr_measure(
            True,
            {f"{measure}_with": better_feature},
            X[feature],
            X[better_feature],
            **params,
        )

        # updating association if it's greater than previous better features
        if association.get(measure_name) > worst_corr.get(measure_name, 0):
            # renaming association measure as filter
            association[f"{measure}_filter"] = association.pop(measure_name)

            # removing temporary measures
            association = {k: v for k, v in association.items() if "_measure" not in k}

            # updating worst known association
            worst_corr.update(association)

        # stopping measurements if association is greater than threshold
        if association.get(f"{measure}_filter") > params.get("thresh_corr", 1):
            ranks = ranks.drop(feature, axis=0)  # removing feature from ranks

            return ranks, None

    return ranks, worst_corr


def qualitative_filter(
    X: DataFrame, ranks: DataFrame, corr_measure: Callable, **params
) -> Dict[str, Any]:
    """Computes max association between X and X (qualitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
        Maximum association between features
    """

    # accessing the prefered order
    prefered_order = ranks.index

    # initiating list of maximum association per feature
    associations = []

    # iterating over each feature by target association order
    for feature in prefered_order:
        # computing correlation with better features anf filtering out ranks
        ranks, worst_corr = qualitative_worst_corr(X, feature, ranks, corr_measure, **params)

        # updating associations
        if worst_corr:
            associations += [worst_corr]

    # formatting ouput to DataFrame
    associations = DataFrame(associations).set_index("feature")

    # applying filter on association
    associations = ranks.join(associations, how="right")

    return associations


def cramerv_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Computes max Cramer's V between X and X (qualitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
        Maximum association between features
    """

    # applying quantitative filter with Cramer's V correlation
    return qualitative_filter(X, ranks, cramerv_measure, **params)


def tschuprowt_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
    """Computes max Tschuprow's T between X and X (qualitative) excluding
     features that are correlated to a feature more associated with the target
    (defined by the ranks).

    Parameters
    ----------
    thresh_corr, float: default 1.
        Maximum association between features
    """

    # applying quantitative filter with Tschuprow's T correlation
    return qualitative_filter(X, ranks, tschuprowt_measure, **params)
